# -*- coding: utf-8 -*-
#  Copyright (c) 2023. Distributed under the terms of the MIT License.

import pytest
from vise.util.structure_handler import sanitize_matrix


def test_sanitize_matrix_9_input_values():
    actual = sanitize_matrix(list(range(9)))
    expected = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    assert actual == expected


def test_sanitize_matrix_3_input_values():
    actual = sanitize_matrix(list(range(1, 4)))
    expected = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
    assert actual == expected


def test_sanitize_matrix_1_input_value():
    actual = sanitize_matrix([4])
    expected = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
    assert actual == expected





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from math import ceil

from pymatgen.core import Composition
from pymatgen.io.vasp import Potcar

from vise.input_set.datasets.dataset_util import (
    LDAU, unoccupied_bands, num_bands, calc_kpar)


def test_ldau_3d_transition_metal():
    actual = LDAU(symbol_list=["Mn", "Cu"], )
    assert actual.ldauu == [3, 5]
    assert actual.ldaul == [2, 2]
    assert actual.lmaxmix == 4


def test_ldau_rare_earth():
    actual = LDAU(symbol_list=["Mn", "Ho"], )
    assert actual.ldauu == [3, 5]
    assert actual.ldaul == [2, 3]
    assert actual.lmaxmix == 6


def test_ldau_3d_override():
    actual = LDAU(symbol_list=["Mn", "Cu"])
    assert actual.ldauu == [3, 5]
    assert actual.ldaul == [2, 2]
    assert actual.lmaxmix == 4


def test_unoccupied_bands():
    assert unoccupied_bands["H"] == 4


def test_nbands():
    composition = Composition("KH2")
    potcar = Potcar(symbols=["Mn", "O"])
    nelect_mn_pv = 7
    nelect_o = 6
    unoccupied_band_mn = 9
    unoccupied_band_o = 4
    expected = (ceil((nelect_mn_pv + nelect_o * 2) / 2)
                + unoccupied_band_mn + unoccupied_band_o * 2)
    assert num_bands(composition, potcar) == expected


def test_calc_kpar():
    # assert calc_kpar(3, num_cores=2, unused_core_ratio_threshold=0.251) == 2
    assert calc_kpar(3, num_cores=2, unused_core_ratio_threshold=0.249) == 1
    assert calc_kpar(5, num_cores=36, unused_core_ratio_threshold=0.249) == 6





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from vise.analyzer.atom_grouping_type import AtomGroupingType


def test_group_by_atoms(complex_monoclinic_structure):
    class_type = AtomGroupingType.atoms
    target = ["0,1", "2,3,4"]
    group = class_type.grouped_atom_indices(structure=complex_monoclinic_structure,
                                            target=target)
    assert group == {"0,1": [0, 1], "2,3,4": [2, 3, 4]}


def test_group_by_elements(complex_monoclinic_structure):
    class_type = AtomGroupingType.elements
    target = ["H", "He"]
    group = class_type.grouped_atom_indices(structure=complex_monoclinic_structure,
                                            target=target)
    assert group == {"H": [0], "He": [1, 2, 3, 4]}


def test_group_by_elements_wo_target(complex_monoclinic_structure):
    class_type = AtomGroupingType.elements
    group = class_type.grouped_atom_indices(structure=complex_monoclinic_structure)
    assert group == {"H": [0], "He": [1, 2, 3, 4]}


def test_group_by_non_equiv_sites(complex_monoclinic_structure):
    class_type = AtomGroupingType.non_equiv_sites
    target = None
    group = class_type.grouped_atom_indices(structure=complex_monoclinic_structure,
                                            target=target)
    assert group == {"H1_a": [0], "He1_m": [1, 2], "He2_m": [3, 4]}


def test(complex_monoclinic_structure):
    print(complex_monoclinic_structure.composition.elements)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pymatgen.core import Element
from vise.cli.main_util_functions import make_atom_poscars, \
    make_spin_decomposed_volumetric_files, make_light_weight_vol_data

_file_path = "vise.cli.main_util_functions"


# def test_make_phonon_poscars():
#     args = Namespace(unitcell=)


def test_make_atom_poscars(mocker):
    args = Namespace(dirname=Path("a"), elements=[Element.H, Element.He])
    mock = mocker.patch(f"{_file_path}.make_atom_poscar_dirs")
    make_atom_poscars(args)
    mock.assert_called_once_with(Path("a"), [Element.H, Element.He])


def test_make_spin_decomposed_volumetric_files(mocker):
    args = Namespace(chgcar="CHGCAR")
    mock_chgcar = mocker.patch(f"{_file_path}.Chgcar")
    mock_make_spin_charges = mocker.patch(f"{_file_path}.make_spin_charges")
    chgcar = mock_chgcar.from_file.return_value
    make_spin_decomposed_volumetric_files(args)
    mock_chgcar.from_file.assert_called_once_with("CHGCAR")
    mock_make_spin_charges.assert_called_once_with(chgcar)


def test_make_light_weight_vol_data(mocker, tmpdir):
    tmpdir.chdir()
    Path("original.vesta").write_text("a")

    args = Namespace(volumetric_file="CHGCAR",
                     output_lw_volumetric_filename=Path("CHGCAR_lw"),
                     border_fractions=[0.1],
                     output_vesta_filename=Path("test_w_chg.vesta"),
                     original_vesta_file=Path("original.vesta"))
    mock_chgcar = mocker.patch(f"{_file_path}.Chgcar")
    mock_vol_text = mocker.patch(f"{_file_path}.light_weight_vol_text")
    mock_calc_isurfs = mocker.patch(f"{_file_path}.calc_isurfs")
    mock_add_density = mocker.patch(f"{_file_path}.add_density")

    chgcar = mock_chgcar.from_file.return_value
    mock_add_density.return_value = "add_density_return"
    mock_vol_text.return_value = "light_weight_vol_text"
    mock_calc_isurfs.return_value = [0.0001]

    make_light_weight_vol_data(args)
    mock_vol_text.assert_called_once_with(chgcar, [0.1])
    mock_calc_isurfs.assert_called_once_with([0.1], True,
                                             chgcar.structure.volume)
    mock_add_density.assert_called_once_with("a", [0.0001], "CHGCAR_lw")
    assert Path("test_w_chg.vesta").read_text() == "add_density_return"
    assert Path("CHGCAR_lw").read_text() == "light_weight_vol_text"



# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from pathlib import Path

from vise.analyzer.vasp.handle_volumetric_data import \
    light_weight_vol_text, make_spin_charges
from pymatgen.core import Structure, Lattice
from pymatgen.io.vasp import Chgcar
import numpy as np
from numpy.testing import assert_almost_equal


def test_make_spin_charges(simple_cubic):
    structure = Structure.from_dict(simple_cubic.as_dict())

    chgcar = Chgcar(structure, data={"total": np.array([[[2.0]]])})
    actual = make_spin_charges(chgcar)
    assert_almost_equal(actual[0].data["total"], np.array([[[1.0]]]))

    chgcar = Chgcar(structure, data={"total": np.array([[[2.0]]]),
                                     "diff": np.array([[2.0]])})
    actual = make_spin_charges(chgcar)
    assert_almost_equal(actual[1].data["total"], np.array([[[0.0]]]))


def test_light_weight_vol_text(tmpdir):
    print(tmpdir)
    tmpdir.chdir()
    structure = Structure(Lattice.cubic(1), species=["O"], coords=[[0]*3])
    chgcar = Chgcar(structure, data={"total": np.array([[[0.0, 0.1, 0.2, -1e-5]]])})
    actual = light_weight_vol_text(volumetric_data=chgcar,
                                   border_fractions=[0.45, 0.55])
    expected = """O1
1.0
   1.0000000000000000    0.0000000000000000    0.0000000000000000
   0.0000000000000000    1.0000000000000000    0.0000000000000000
   0.0000000000000000    0.0000000000000000    1.0000000000000000
O
1
direct
   0.0000000000000000    0.0000000000000000    0.0000000000000000 O

1 1 4
0 1 2 0"""
    assert actual == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from vise.input_set.datasets.potcar_set import PotcarSet


def test_potcar_list_normal():
    potcar_list = PotcarSet.normal.potcar_dict()
    assert potcar_list["H"] == "H"


def test_potcar_list_gw():
    potcar_list = PotcarSet.gw.potcar_dict()
    assert potcar_list["Li"] == "Li_GW"


def test_potcar_list_non():
    potcar_list = PotcarSet.mp_relax_set.potcar_dict()
    assert potcar_list["Fr"] is None
    assert potcar_list["Sr"] == "Sr_sv"


def test_potcar_set():
    assert PotcarSet.normal.overridden_potcar_dict()["Li"] == "Li"


def test_potcar_set_override():
    actual = PotcarSet.normal.overridden_potcar_dict({"Li": "Li_test"})["Li"]
    assert actual == "Li_test"


# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from copy import copy
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
import pytest
from monty.json import MSONable, MontyDecoder
from monty.serialization import loadfn
from pymatgen.core import IStructure, Lattice, Structure
from vise.tests.helpers.assertion import assert_dataclass_almost_equal, \
    assert_structure_almost_same, assert_msonable, assert_yaml_roundtrip
from vise.util.mix_in import ToYamlFileMixIn

decoder = MontyDecoder()


@dataclass
class Data:
    x: float


def test_assert_msonable_round_trip_cover_all_attributes():
    @dataclass
    class TestMSONableData(MSONable):
        x: float = None
        y: str = None

        def as_dict(self):
            return {"x": self.x}

        @classmethod
        def from_dict(cls, d):
            kwargs = copy(d)
            kwargs.pop("@module", None)
            kwargs.pop("@class", None)
            return cls(**d)

    with pytest.raises(AssertionError):
        assert_msonable(TestMSONableData(1.0, "a"))


@dataclass
class YamlRoundTripClass(ToYamlFileMixIn):
    structures: List[Structure] = None

    def as_dict(self):
        return {"structures": [self.structures[0].as_dict()]}

    @classmethod
    def from_dict(cls, d):
        return cls(structures=[decoder.process_decoded(d["structures"][0])])


def test_assert_yaml_roundtrip(tmpdir):
    coords = [[0.0, 0.0, 0.0]]
    s1 = Structure(lattice=Lattice.cubic(1.0), species=["H"], coords=coords)
    yaml_class = YamlRoundTripClass(structures=[s1])
    expected = """structures:
- '@class': Structure
  '@module': pymatgen.core.structure
  charge: 0
  lattice:
    a: 1.0
    alpha: 90.0
    b: 1.0
    beta: 90.0
    c: 1.0
    gamma: 90.0
    matrix:
    - - 1.0
      - 0.0
      - 0.0
    - - 0.0
      - 1.0
      - 0.0
    - - 0.0
      - 0.0
      - 1.0
    volume: 1.0
  sites:
  - abc:
    - 0.0
    - 0.0
    - 0.0
    label: H
    properties: {}
    species:
    - element: H
      occu: 1
    xyz:
    - 0.0
    - 0.0
    - 0.0
"""
    assert_yaml_roundtrip(yaml_class, tmpdir, expected)


@dataclass
class TestData:
    a: float = None
    b: List[float] = None
    c: List[Tuple[float]] = None
    d: Data = None


def test_assert_structure_almost_same():
    coords = [[0.0, 0.0, 0.0]]
    s1 = IStructure(lattice=Lattice.cubic(1.0), species=["H"], coords=coords)
    coords = [[0.0, 0.0, 0.0]]

    s2 = IStructure(lattice=Lattice.cubic(1.000001), species=["H"],
                    coords=coords)
    assert_structure_almost_same(s1, s2)

    s3 = IStructure(lattice=Lattice.cubic(1.001), species=["H"], coords=coords)
    with pytest.raises(AssertionError):
        assert_structure_almost_same(s1, s3)


def test_assert_dataclass_almost_equal_float():
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), 1.0)

    assert_dataclass_almost_equal(TestData(1.0), TestData(1.0))

    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), TestData(2.0))

    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), TestData(1.01), digit=2)

    assert_dataclass_almost_equal(TestData(1.0), TestData(1.01), digit=1)


def test_assert_dataclass_almost_equal_w_different_type_attr():
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(1.0), TestData(np.float32(1.0)))


def test_assert_dataclass_almost_equal_iterable():
    assert_dataclass_almost_equal(TestData(b=[1.0]), TestData(b=[1.0]))

    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(b=[1.0]), TestData(b=[2.0]))


def test_assert_dataclass_almost_equal_nested_iterable():
    assert_dataclass_almost_equal(TestData(c=[(1.0,)]), TestData(c=[(1.0,)]))
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(c=[(1.0,)]),
                                      TestData(c=[(2.0,)]), digit=2)


def test_assert_dataclass_almost_equal_dataclass_attribute():
    assert_dataclass_almost_equal(TestData(d=Data(1.0)), TestData(d=Data(1.0)))
    with pytest.raises(AssertionError):
        assert_dataclass_almost_equal(TestData(d=Data(1.0)), TestData(d=Data(1.01)), digit=2)

    assert_dataclass_almost_equal(TestData(d=Data(1.0)), TestData(d=Data(1.001)), digit=2)





# -*- coding: utf-8 -*-
#  Copyright (c) 2023. Distributed under the terms of the MIT License.

from dataclasses import dataclass
from typing import List, Dict

from monty.json import MSONable


@dataclass
class BZPlotInfo(MSONable):
    faces: List[List[List[float]]]
    # {"X": {"cart": [0.5, 0, 0], "frac": [0.7514, 0, 0]"}}
    labels: Dict[str, Dict[str, List[float]]]
    band_paths: List[List[List[float]]] = None
    rec_lat_vec: List[List[float]] = None



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import matplotlib.pyplot as plt

from vise.util.matplotlib import my_formatter, float_to_int_formatter


def test_formatter():
    assert my_formatter(0.49e-10, None) == 0
    assert my_formatter(0.51e-10, None) == 1e-10
    assert my_formatter("a", None) == "a"


def test_plot():
    plt.xlim(0, 1)
    plt.ylim(0, 1)
    axis = plt.gca()
    axis.xaxis.set_major_formatter(float_to_int_formatter)
    plt.show()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Tuple, Dict

import numpy as np
import seekpath
import spglib
from monty.json import MSONable
from more_itertools import consecutive_groups
from pymatgen.core import Structure, Element
from collections import defaultdict
from itertools import groupby

from spglib import SpglibDataset
from tabulate import tabulate
from vise.defaults import defaults
from vise.util.centering import Centering
from vise.util.logger import get_logger
from vise.error import ViseError
from vise.util.bravais_lattice import BravaisLattice

logger = get_logger(__name__)


def cell_to_structure(
        cell: Tuple[List[List[float]], List[List[float]], List[int]]
) -> Structure:
    """
    cell: (Lattice parameters
           [[a_x, a_y, a_z], [b_x, b_y, b_z], [c_x, c_y, c_z]],
          Fractional atomic coordinates in an Nx3 array,
          Z numbers of species in a length N array)
    """
    species = [Element.from_Z(i) for i in cell[2]]
    return Structure(lattice=cell[0], coords=cell[1], species=species)


class StructureSymmetrizer:
    def __init__(self,
                 structure: Structure,
                 symprec: float = defaults.symmetry_length_tolerance,
                 angle_tolerance: float = defaults.symmetry_angle_tolerance,
                 time_reversal: bool = True,
                 band_mesh_distance: float = defaults.band_mesh_distance):
        """Get full information of seekpath band path.

        Note: site properties such as magmom are removed.

        The structures of aP (SG:1, 2), mC (5, 8, 9, 12, 15) and
        oA (38, 39, 40, 41) can be different between spglib and seekpath.
        see Y. Hinuma et al. Comput. Mater. Sci. 128 (2017) 140–184
        -- spglib mC
         6.048759 -3.479491 0.000000
         6.048759  3.479491 0.000000
        -4.030758  0.000000 6.044512
        -- seekpath mC
         6.048759  3.479491  0.000000
        -6.048759  3.479491  0.000000
        -4.030758  0.000000  6.044512
        """
        self.structure = structure.copy()
        if structure.site_properties:
            logger.warning(f"Site property {structure.site_properties.keys()} "
                           f"removed in primitive and conventional structures.")
        self.symprec = symprec
        self.angle_tolerance = angle_tolerance
        self.time_reversal = time_reversal
        self.ref_distance = band_mesh_distance
        lattice_matrix = structure.lattice.matrix
        positions = structure.frac_coords.tolist()
        atomic_numbers = [i.specie.number for i in structure.sites]
        self.cell = (lattice_matrix, positions, atomic_numbers)
        # evaluated lazily
        self._spglib_sym_data = None
        self._primitive = None
        self._second_primitive = None
        self._seekpath_data = None
        self._band_primitive = None

    def __repr__(self):
        sym_data = self.spglib_sym_data
        is_primitive = self.structure == self.primitive
        lines = [f"Symprec: {self.symprec}",
                 f"Angle tolerance: {self.angle_tolerance}",
                 f"Space group: {sym_data['international']}",
                 f"Is primitive: {is_primitive}"]
        site_info_header = ["site", "wyckoff", "site sym", "equiv sites"]
        site_infos = []
        for name, site in self.sites.items():
            site_infos.append([f"{name}",
                               f"{site.wyckoff_letter}",
                               f"{site.site_symmetry}",
                               f"{site.pprint_equiv_atoms}"])
        lines.append(tabulate(site_infos, headers=site_info_header))

        return "\n".join(lines)

    @property
    def spglib_sym_data(self) -> SpglibDataset:
        if not self._spglib_sym_data:
            self._spglib_sym_data = spglib.get_symmetry_dataset(
                self.cell, self.symprec, self.angle_tolerance)
        return self._spglib_sym_data

    @property
    def conventional(self) -> Structure:
        center = Centering.from_string(self.centering)
        return self.primitive * center.primitive_to_conv

        # # I don't know if this is fine for spglib cyclic behavior.
        # if self._conventional is None:
        #     conventional = \
        #         spglib.standardize_cell(self.cell,
        #                                 symprec=self.symprec,
        #                                 angle_tolerance=self.angle_tolerance)
        #     if conventional is None:
        #         raise ViseSymmetryError(
        #             "Spglib couldn't find the conventional cell. Change the "
        #             "symprec and/or angle_tolerance.")
        #     else:
        #         self._conventional = \
        #             cell_to_structure(conventional).get_sorted_structure()
        # return self._conventional

    @property
    def primitive(self) -> Structure:
        if self._primitive is None:
            primitive = \
                spglib.find_primitive(self.cell,
                                      symprec=self.symprec,
                                      angle_tolerance=self.angle_tolerance)
            if primitive is None:
                raise ViseSymmetryError(
                    "Spglib couldn't find the primitive cell. "
                    "Change the symprec and/or angle_tolerance.")
            else:
                # To manage spglib cyclic behavior, we need to run this again.
                second_primitive = cell_to_structure(spglib.find_primitive(
                    primitive, symprec=self.symprec,
                    angle_tolerance=self.angle_tolerance)).get_sorted_structure()
                primitive = cell_to_structure(primitive)

                if primitive != second_primitive:
                    if first_structure_is_primitive(primitive, second_primitive):
                        self._primitive = primitive
                        self._second_primitive = second_primitive

                    else:
                        self._primitive = second_primitive
                        self._second_primitive = primitive
                else:
                    self._primitive = primitive

        return self._primitive

    @property
    def second_primitive(self) -> Structure:
        return self._second_primitive

    def find_seekpath_data(self) -> None:
        """Get full information of seekpath band path. """
        logger.info(f"Band mesh distance is set to {self.ref_distance}. "
                    f"To change it, use band_ref_dist option.")
        self._seekpath_data = \
            seekpath.get_explicit_k_path(structure=self.cell,
                                         symprec=self.symprec,
                                         angle_tolerance=self.angle_tolerance,
                                         with_time_reversal=self.time_reversal,
                                         reference_distance=self.ref_distance)
        lattice = self._seekpath_data["primitive_lattice"]
        element_types = self._seekpath_data["primitive_types"]
        species = [Element.from_Z(i) for i in element_types]
        positions = self._seekpath_data["primitive_positions"]
        self._band_primitive = Structure(lattice, species, positions)

    @property
    def sg_number(self):
        return self.spglib_sym_data.number

    @property
    def space_group(self):
        return self.spglib_sym_data.international

    @property
    def point_group(self):
        return self.spglib_sym_data.pointgroup

    @property
    def seekpath_data(self):
        if self._seekpath_data is None:
            self.find_seekpath_data()
        return self._seekpath_data

    @property
    def band_primitive(self) -> Structure:
        if self._band_primitive is None:
            self.find_seekpath_data()
        return self._band_primitive

    @property
    def is_primitive_lattice_changed(self) -> bool:
        # np.allclose is used for lattice comparison.
        # def allclose(a, b, rtol=1.e-5, atol=1.e-8, equal_nan=False):
        return self.structure.lattice != self.primitive.lattice

    @property
    def is_band_primitive_lattice_changed(self) -> bool:
        return self.structure.lattice != self.band_primitive.lattice

    @property
    def band_primitive_differ_primitive(self) -> bool:
        return self.primitive.lattice != self.band_primitive.lattice

    def irreducible_kpoints(self,
                            num_kpt_list: List[int],
                            kpt_shift: List[float]  # vasp convention
                            ) -> List[Tuple[List[float], int]]:
        # evaluated at a given Brillouin zone.
        mapping, integer_grid_points = spglib.get_ir_reciprocal_mesh(
            mesh=num_kpt_list,
            cell=self.cell,
            is_shift=np.array(kpt_shift) * 2,
            is_time_reversal=self.time_reversal,
            symprec=self.symprec)

        results = []
        shift_in_integer_grid = np.array(kpt_shift)
        for repr_index, count in zip(*np.unique(mapping, return_counts=True)):
            integer_grid_point = \
                integer_grid_points[repr_index] + shift_in_integer_grid
            normalized_grid_point = integer_grid_point / num_kpt_list
            results.append((normalized_grid_point.tolist(), count))

        return results  # [(irreducible kpoint in fractional coords, weight)]

    def grouped_atom_indices(self):
        result = {}
        for name, site in self.sites.items():
            key = "_".join([name, site.wyckoff_letter])
            result[key] = site.equivalent_atoms
        return result

    @property
    def sites(self) -> Dict[str, "Site"]:
        wyckoffs = self.spglib_sym_data.wyckoffs
        equivalent_atoms = self.spglib_sym_data.equivalent_atoms
        site_symmetries = self.spglib_sym_data.site_symmetry_symbols
        equiv_indices = sorted(enumerate(equivalent_atoms), key=lambda x: x[1])
        result = {}
        element_idx_dict = defaultdict(int)
        for _, equiv_sites in groupby(equiv_indices, lambda x: x[1]):
            equiv_site_list = list(equiv_sites)
            repr_idx = equiv_site_list[0][0]
            element = self.structure[repr_idx].specie.name
            element_idx_dict[element] += 1
            index = str(element_idx_dict[str(element)])
            name = element + index
            result[name] = Site(element=element,
                                wyckoff_letter=wyckoffs[repr_idx],
                                site_symmetry=site_symmetries[repr_idx],
                                equivalent_atoms=[s[0] for s in equiv_site_list])
        return result

    @property
    def bravais(self):
        return BravaisLattice.from_sg_num(self.spglib_sym_data.number)

    @property
    def centering(self):
        return self.spglib_sym_data.international[0]


def first_structure_is_primitive(structure1: Structure, structure2: Structure):
    for s1, s2 in zip(structure1, structure2):
        for c1, c2 in zip(s1.frac_coords, s2.frac_coords):
            if c1 < c2 - 1e-5:
                return True
            elif c2 < c1 - 1e-5:
                return False
    raise ViseSymmetryError("Primitive cannot be unique.")


@dataclass(frozen=True)
class Site(MSONable):
    element: str
    wyckoff_letter: str
    site_symmetry: str
    equivalent_atoms: List[int]

    @property
    def pprint_equiv_atoms(self):
        str_list = []
        for consecutive_ints in consecutive_groups(self.equivalent_atoms):
            ints = list(consecutive_ints)
            if len(ints) >= 3:
                str_list.append("..".join([str(ints[0]), str(ints[-1])]))
            else:
                str_list.append(" ".join([str(j) for j in ints]))
        return " ".join(str_list)


num_sym_op = {"1":     1,
              "-1":    2,
              "2":     2,
              "m":     2,
              "2/m":   4,
              "222":   4,
              "2mm":   4,
              "m2m":   4,
              "mm2":   4,
              "mmm":   8,
              "4":     4,
              "-4":    4,
              "4/m":   8,
              "422":   8,
              "4mm":   8,
              "-4m2":  8,
              "-42m":  8,
              "4/mmm": 16,
              "3":     3,
              "-3":    6,
              "32":    6,
              "3m":    6,
              "-3m":   12,
              "6":     6,
              "-6":    6,
              "6/m":   12,
              "622":   12,
              "6mm":   12,
              "-6m2":  12,
              "6/mmm": 24,
              "23":    12,
              "m3":    24,
              "m-3":   24,
              "432":   24,
              "-43m":  24,
              "m-3m":  48}


def num_symmetry_operation(point_group: str) -> int:
    """ Return number of symmetry operations from Hermann–Mauguin notation.

    Args:
         point_group (str):
            Point group in Hermann–Mauguin notation.
            "." will be removed, e.g., ..6 -> 6.

    Returns:
         Number of symmetry operation.
    """
    # remove "." from the given point_group
    point_group = "".join([s for s in point_group if s != "."])
    return num_sym_op[point_group]


class ViseSymmetryError(ViseError):
    pass



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from abc import abstractmethod

from typing import List

import numpy as np
from matplotlib import pyplot as plt
from vise.analyzer.dielectric_function import DieleFuncData
from vise.util.enum import ExtendedEnum
from vise.util.matplotlib import float_to_int_formatter


class TensorDirection(ExtendedEnum):
    average = "average"
    xx = "xx"
    yy = "yy"
    zz = "zz"
    xy = "xy"
    yz = "yz"
    xz = "xz"

    def val(self, tensors: List[List[float]]) -> List[float]:
        if self is self.average:
            return [(t[0] + t[1] + t[2]) / 3 for t in tensors]
        else:
            idx = ["xx", "yy", "zz", "xy", "yz", "xz"].index(self.name)
            return [tensor[idx] for tensor in tensors]


class DieleFuncPlotType(ExtendedEnum):
    diele_func = "diele_func"
    absorption_coeff = "absorption_coeff"
    refraction = "refraction"
    reflectivity = "reflectivity"

    def y_axis_label(self, plot_engine):
        if self is self.diele_func:
            label = "Dielectric function"
        elif self is self.absorption_coeff:
            label = "Absorption coefficient"
        elif self is self.refraction:
            label = "Refraction"
        elif self is self.reflectivity:
            label = "Reflectivity"
        else:
            raise ValueError(f"Need to implement label for {self}.")

        if self is self.absorption_coeff:
            if plot_engine == "plotly":
                sup_before, sup_after = "<sup>", "</sup>"
            elif plot_engine == "matplotlib":
                sup_before, sup_after = "$^{", "}$"
            else:
                raise ValueError("The plotter engine is not adequate.")
            return f"{label} (cm{sup_before}-1{sup_after})"

        return label

    @property
    def y_axis_default_min(self):
        if self is self.absorption_coeff:
            return 10**3, 10**7
        return None

    def tensors(self, data: DieleFuncData):
        if self is self.diele_func:
            return data.diele_func_real, data.diele_func_imag
        elif self is self.absorption_coeff:
            return data.absorption_coeff,
        elif self is self.refraction:
            return data.refractive_idx_real, data.refractive_idx_imag
        elif self is self.reflectivity:
            return data.reflectivity,


def auto_y_range(plot_type, values):
    _max = max(values)
    if plot_type is DieleFuncPlotType.absorption_coeff:
        y_min, y_max = 10 ** 3, _max * 2
    else:
        _min = min(values)
        y_min, y_max = 1.05 * _min - .05 * _max, 1.05 * _max - .05 * _min
    return y_min, y_max


class DieleFuncPlotter:

    def __init__(self, diele_func_data: DieleFuncData,
                 energy_range: List[float] = None):

        self.diele_func_data = diele_func_data
        self.energies = diele_func_data.energies
        self.band_gap = diele_func_data.band_gap
        self.energy_range = energy_range or [0, 10]
        self._x_axis_title = "Energy (eV)"

    def add_plot(self, directions, plot_type: DieleFuncPlotType):
        _idx = np.where(np.array(self.energies) > self.energy_range[1])[0][0]
        values = []
        a = plot_type.tensors(self.diele_func_data)
        for b, tensors in zip(["real", "imag"], a):
            for direction in directions:
                _index = self.diele_func_data.directions.index(direction)
                tensor = tensors[_index]
                values.extend(tensor[:_idx])
                if len(a) == 1:
                    name = direction
                else:
                    name = f"{b}_{direction}"
                self.add_single_plot(name, tensor, plot_type)
        return values

    @abstractmethod
    def add_single_plot(self, name, tensor, plot_type):
        pass


class DieleFuncMplPlotter(DieleFuncPlotter):
    def __init__(self, diele_func_data: DieleFuncData,
                 energy_range: List[float] = None):
        super().__init__(diele_func_data, energy_range)
        self.title = diele_func_data.title
        self.plt = plt
        self.plt.clf()

    def construct_plot(self,
                       directions=("ave",),
                       plot_type=DieleFuncPlotType.absorption_coeff,
                       y_range=None):
        self._add_coeffs(directions, plot_type, y_range)
        self._add_band_gap()
        self._set_figure_legend()
        self._set_x_range()
        self._set_labels(plot_type)
        self._set_formatter()
        self._set_title()
        self.plt.tight_layout()

    def _add_coeffs(self, directions, plot_type, y_range):
        all_plotted_values = self.add_plot(directions, plot_type)
        y_min, y_max = y_range or auto_y_range(plot_type, all_plotted_values)
        self.plt.gca().set_ylim(ymin=y_min, ymax=y_max)

    def add_single_plot(self, name, tensor, plot_type):
        if plot_type is DieleFuncPlotType.absorption_coeff:
            self.plt.semilogy(self.energies, tensor, label=name)
        else:
            self.plt.plot(self.energies, tensor, label=name)

    def _add_band_gap(self):
        self.plt.axvline(x=self.band_gap, linestyle="dashed", color="black",
                         linewidth=1)

    def _set_figure_legend(self):
        self.plt.legend()

    def _set_x_range(self):
        self.plt.xlim(self.energy_range[0], self.energy_range[1])

    def _set_labels(self, plot_type):
        self.plt.xlabel(self._x_axis_title)
        y_axis = plot_type.y_axis_label("matplotlib")
        self.plt.ylabel(y_axis)

    def _set_formatter(self):
        axis = self.plt.gca()
        axis.xaxis.set_major_formatter(float_to_int_formatter)

    def _set_title(self):
        if self.title:
            self.plt.title(self.title)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from vise.util.enum import ExtendedEnum


class KpointsMode(ExtendedEnum):
    """K-point generation type

       Supporting modes are:
       "band":
           Kpoints with the band path will be returned based on the
           seekpath program. The space group is analyzed and primitive
           unitcell that must be used for the band structure calculation is
           returned as well.
       "primitive":
           Kpoints with uniform k-point sampling. The k-point sampling mesh
           and centering are determined based on the standardized primitive
           unitcell. Structure is also changed if not primitive.
       "uniform":
           Kpoints with uniform k-point sampling. The k-point sampling mesh
           and centering are determined based on the given lattice. Note
           that only when the angles are 90 degrees, the centering is
           shifted along the perpendicular direction.
           This mode is useful when calculating the supercells.
    """
    band = "band"
    primitive = "primitive"
    uniform = "uniform"

    @property
    def band_or_primitive(self):
        return True if self in [self.band, self.primitive] else False



# -*- coding: utf-8 -*-

#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from copy import deepcopy
from pathlib import Path
from typing import Optional, Dict

from monty.serialization import loadfn

from vise.util.enum import ExtendedEnum


class PotcarSet(ExtendedEnum):
    normal = "normal"
    mp_relax_set = "mp_relax_set"
    gw = "gw"

    def overridden_potcar_dict(self,
                               override_potcar_set: Optional[dict] = None
                               ) -> Dict[str, str]:
        result = deepcopy(self.potcar_dict())
        if override_potcar_set:
            result.update(override_potcar_set)

        return result  # e.g. {"Zr": "Zr_pv", ...}

    def potcar_dict(self) -> Dict[str, str]:
        _potcar_list = loadfn(Path(__file__).parent / "potcar_set.yaml")
        set_names = _potcar_list.pop("set_names")
        result = {}
        for name in set_names:
            result[name] = {}  # normal, mp_relax_set, gw

        for element, potcar_string in _potcar_list.items():
            potcars = potcar_string.split()

            def sanitize(val: str):
                if val == "---":
                    return potcars[0]
                elif val == "None":
                    return None
                else:
                    return val

            for index, potcar_single in enumerate(potcars):
                set_name = set_names[index]
                result[set_name][element] = sanitize(potcar_single)

        return result[self.value]


# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from typing import Tuple, List, Union

Coords = Tuple[float, float, float]
GenCoords = Union[Tuple[float, float, float], List[float]]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from typing import List, Optional

import matplotlib.pyplot as plt

from vise.analyzer.dos_data import DosPlotData
from vise.util.matplotlib import float_to_int_formatter


class DosMplSettings:
    def __init__(self,
                 colors: Optional[List[str]] = None,
                 linewidth: float = 1.0,
                 band_edge_line_width: float = 0.75,
                 band_edge_line_color: str = "black",
                 band_edge_line_style: str = "-.",
                 title_font_size: int = 15,
                 label_font_size: int = 12,
                 ):
        self.colors = colors or ['#36454f', '#E15759', '#4E79A7', '#F28E2B']
        self.linewidth = linewidth

        self.vline = {"linewidth": band_edge_line_width,
                      "color": band_edge_line_color,
                      "linestyle": band_edge_line_style}

        self.title_font_size = title_font_size
        self.label_font_size = label_font_size

    def dos_line(self, index):
        return {"color": self.colors[index],
                "linewidth": self.linewidth}


class DosPlotter:
    def __init__(self,
                 dos_plot_data: DosPlotData,
                 show_legend: bool = True,
                 mpl_defaults: Optional[DosMplSettings] = DosMplSettings()):

        self._dos_plot_data = dos_plot_data
        self._show_legend = show_legend
        self.mpl_defaults = mpl_defaults
        self.title = dos_plot_data.title
        self.plt = plt

        num_axs = len(self._dos_plot_data.doses)
        self.fig, self._axs = self.plt.subplots(num_axs, 1,
                                                sharex=True,
                                                gridspec_kw={'hspace': 0.1})
        if num_axs == 1:
            self._axs = [self._axs]

    def construct_plot(self):
        self._axs[0].set_xlim(self._dos_plot_data.energy_range)
        self._axs[0].xaxis.set_major_formatter(float_to_int_formatter)
        for i in range(len(self._dos_plot_data.doses)):
            self._add_ax(i)
        self._set_x_labels()
        middle_idx = int(i / 2)
        self._set_ylabel(middle_idx)
        self._set_title()

    def _add_ax(self, i):
        self._add_dos(i)
        self._set_y_range(i)
        self._set_dos_zero_line(i)
        self._set_vline(i)
        self._set_formatter(i)

    def _add_dos(self, i):
        for j, by_name_dos in enumerate(self._dos_plot_data.doses[i]):
            for k, by_spin_dos in enumerate(by_name_dos.dos):
                sign = - (2 * k) + 1  # up: 1, down: -1
                dos_for_plot = [d * sign for d in by_spin_dos]
                args = self.mpl_defaults.dos_line(j)
                if k == 0 and self._show_legend:
                    if self._dos_plot_data.doses[i][j].name:
                        args["label"] = "-".join(
                            [self._dos_plot_data.names[i],
                             self._dos_plot_data.doses[i][j].name])
                    else:
                        args["label"] = self._dos_plot_data.names[i]
                self._axs[i].plot(self._dos_plot_data.relative_energies,
                                  dos_for_plot, **args)

        if self._show_legend:
            self._axs[i].legend(bbox_to_anchor=(0.9, 1), loc='upper left',
                                borderaxespad=0, markerscale=0.1)

    def _set_ylabel(self, i):
        self._axs[i].set_ylabel("Dos (1/eV)",
                                size=self.mpl_defaults.label_font_size)

    def _set_dos_zero_line(self, i):
        self._axs[i].axhline(0, linestyle=":", color="black")

    def _set_y_range(self, i):
        self._axs[i].set_ylim(self._dos_plot_data.dos_ranges[i])

    def _set_vline(self, i):
        for line_energy in self._dos_plot_data.energy_lines:
            self._axs[i].axvline(x=line_energy, **self.mpl_defaults.vline)

    def _set_formatter(self, i):
        self._axs[i].yaxis.set_major_formatter(float_to_int_formatter)

    def _set_x_labels(self):
        self.plt.xlabel("Energy (eV)", size=self.mpl_defaults.label_font_size)

    def _set_title(self):
        if self.title:
            if self.fig:
                self.fig.suptitle(self.title,
                                  fontsize=self.mpl_defaults.title_font_size)
                self.fig.subplots_adjust(top=0.92)
            else:
                self.plt.title(self.title,
                               fontsize=self.mpl_defaults.title_font_size)


# -*- coding: utf-8 -*-
#  Copyright (c) 2022. Distributed under the terms of the MIT License.


import pytest
from vise.input_set.datasets.potcar_set import PotcarSet
from vise.input_set.kpoints_mode import KpointsMode

from vise.input_set.task import Task
from vise.input_set.vise_log import ViseLog
from vise.input_set.xc import Xc
from vise.tests.helpers.assertion import assert_yaml_roundtrip


@pytest.fixture
def vise_log():
    return ViseLog(version="0.6.3", task=Task.structure_opt, xc=Xc.pbe,
                   input_options={"charge": 1,
                                  "kpt_mode": KpointsMode.uniform,
                                  "potcar_set": PotcarSet.mp_relax_set},
                   user_incar_settings={"NSW": 2},
                   ldauu={"Mg": 3}, ldaul={"Mg": 4})


def test_vise_log(vise_log, tmpdir):
    expected_text = """version: 0.6.3
task: structure_opt
xc: pbe
input_options:
  charge: 1
  kpt_mode: uniform
  potcar_set: mp_relax_set
user_incar_settings:
  NSW: 2
ldauu:
  Mg: 3
ldaul:
  Mg: 4
"""
    assert_yaml_roundtrip(vise_log, tmpdir, expected_text)


def test_vise_log_no_incar_settings(tmpdir):
    vise_log_no_incar_settings = ViseLog(version="0.6.3",
                                         task=Task.structure_opt, xc=Xc.pbe,
                                         input_options={"charge": 1})
    expected_text = """version: 0.6.3
task: structure_opt
xc: pbe
input_options:
  charge: 1
"""
    assert_yaml_roundtrip(vise_log_no_incar_settings, tmpdir, expected_text)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from vise.input_set.incar import ViseIncar, tabulated_string


def test_from_string():
    incar_string = """
    ENCUT = 500; LREAD = True
     ALGO = D     ; PREC = A  
      MAGMOM = 5 5 5
    """
    actual = ViseIncar.from_string(incar_string)
    expected = {'ALGO': 'D', 'ENCUT': 500, 'LREAD': True,
                'MAGMOM': [5, 5, 5], 'PREC': 'A'}
    assert actual == expected


def test_get_string():
    incar = ViseIncar.from_dict({"PREC": "Normal",
                                 "LREAL": False,
                                 "EDIFF": 1.0,
                                 "ISYM": 0,
                                 "MAGMOM": [3, 3, 3],
                                 "PHON_NSTRUCT": 2})
    print(incar)
    actual = incar.get_string()
    expected = """# accuracy
PREC   =  Normal
LREAL  =  False
EDIFF  =  1.0

# symmetry
ISYM  =  0

# spin
MAGMOM  =  3 3 3

# others
PHON_NSTRUCT  =  2"""
    assert actual == expected


def test_setting_to_str():
    incar = ViseIncar.from_dict({"EDIFF": 1.0,
                                 "MAGMOM": [3, 3, 3]})
    assert incar.setting_to_str("EDIFF") == str(1.0)
    assert incar.setting_to_str("MAGMOM") == "3 3 3"


def test_is_ncs_calc():
    incar = ViseIncar.from_dict({"LNONCOLLINEAR": "True"})
    assert incar.is_ncl_calc
    incar = ViseIncar.from_dict({})
    assert incar.is_ncl_calc is False


def test_tabulated_string():
    actual = tabulated_string([["short", "short"], ["long_str", "long_str"]])
    expected = """short     =  short
long_str  =  long_str"""
    assert actual == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import copy

import pytest
from monty.serialization import loadfn
from vise.analyzer.plot_diele_func_data import (DieleFuncMplPlotter,
                                                DieleFuncPlotType)


def test_diele_func_plot_type():
    actual = DieleFuncPlotType.absorption_coeff.y_axis_label("plotly")
    expected = "Absorption coefficient (cm<sup>-1</sup>)"
    assert actual == expected
    actual = DieleFuncPlotType.absorption_coeff.y_axis_label("matplotlib")
    expected = "Absorption coefficient (cm$^{-1}$)"
    assert actual == expected


@pytest.fixture
def actual_diele_func_data(test_data_files):
    result = loadfn(test_data_files / "MgSe_diele_func_data.json")
    result.title = "test diele"
    return result


def test_absorption_coeff_mpl_plotter(actual_diele_func_data):
    plotter = DieleFuncMplPlotter(actual_diele_func_data)
    plotter.construct_plot()
    plotter.plt.show()


def test_refraction_mpl_plotter(actual_diele_func_data):
    plotter = DieleFuncMplPlotter(actual_diele_func_data)
    plotter.construct_plot(plot_type=DieleFuncPlotType.refraction)
    plotter.plt.show()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import pytest
from pymatgen.core import Lattice, IStructure
from vise.util.centering import Centering
from vise.util.structure_symmetrizer import StructureSymmetrizer


def test_names():
    assert Centering.names_string() == "P, A, C, R, I, F"


@pytest.fixture(scope="session")
def rhombohedral():
    lattice = Lattice.rhombohedral(a=1, alpha=45)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def a_centered_orthorhombic():
    lattice = Lattice([[1,  0, 0],
                       [0,  2, 3],
                       [0, -2, 3]])
    coords = [[0.5, 0.8, 0.8],
              [0.0, 0.3, 0.0],
              [0.0, 0.0, 0.3]]

    return IStructure(lattice=lattice, species=["H"] * 3, coords=coords)


@pytest.fixture(scope="session")
def c_centered_monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]]
    return IStructure(lattice=lattice, species=["H", "H"], coords=coords)


@pytest.fixture(scope="session")
def bcc():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 2, coords=coords)


def test_a_centering(a_centered_orthorhombic):
    s = StructureSymmetrizer(a_centered_orthorhombic)
    primitive = s.primitive
    to_conventional = primitive * Centering.A.primitive_to_conv
    assert to_conventional.lattice == s.conventional.lattice
    assert Centering.A.conv_multiplicity == 2


def test_c_centering(c_centered_monoclinic):
    s = StructureSymmetrizer(c_centered_monoclinic)
    primitive = s.primitive
    to_conventional = primitive * Centering.C.primitive_to_conv
    assert to_conventional == s.conventional
    assert Centering.C.conv_multiplicity == 2


def test_r_centering(rhombohedral):
    s = StructureSymmetrizer(rhombohedral)
    rhombohedral = s.primitive
    to_conventional = rhombohedral * Centering.R.primitive_to_conv
    assert to_conventional == s.conventional
#    assert Centering.R.conv_multiplicity == 3


def test_f_centering(bcc):
    s = StructureSymmetrizer(bcc)
    primitive = s.primitive
    to_conventional = primitive * Centering.I.primitive_to_conv
    assert to_conventional == s.conventional
    assert Centering.I.conv_multiplicity == 2


"""
TODO
- 

DONE
"""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import argparse
import sys
import warnings
from pathlib import Path

from monty.serialization import loadfn
from pymatgen.core import Element, Structure
from pymatgen.io.vasp.inputs import UnknownPotcarWarning
from vise.analyzer.vasp.handle_volumetric_data import default_border_fractions
from vise.cli.main import description, epilog
from vise.cli.main_util_functions import make_atom_poscars, \
    make_spin_decomposed_volumetric_files, make_light_weight_vol_data, \
    make_phonon_poscars, make_phonon_figs
from vise.util.logger import get_logger

logger = get_logger(__name__)


warnings.simplefilter('ignore', UnknownPotcarWarning)


def parse_args(args):
    parser = argparse.ArgumentParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    subparsers = parser.add_subparsers()

    # -- make_phonon_poscars ---------------------------------------------------
    parser_make_phonon_poscars = subparsers.add_parser(
        name="make_phonon_poscars",
        description="Tools for .",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['mpp'])

    parser_make_phonon_poscars.add_argument(
        "-u", "--unitcell", type=Structure.from_file,
        help="")
    parser_make_phonon_poscars.add_argument(
        "-s", "--supercell_matrix", type=int, nargs="+",
        help="Supercell matrix for phonon calculation"
             "based on the conventional cell.")

    parser_make_phonon_poscars.set_defaults(func=make_phonon_poscars)

    # -- make_phonon_figs ---------------------------------------------------
    parser_make_phonon_figs = subparsers.add_parser(
        name="make_phonon_figs",
        description="Tools for .",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['mpf'])

    parser_make_phonon_figs.add_argument(
        "-pi", "--phonopy_input", type=loadfn, help="")
    parser_make_phonon_figs.add_argument(
        "-vn", "--vasprun_name", type=str, help="")

    parser_make_phonon_figs.set_defaults(func=make_phonon_figs)

    # -- make_atom_poscars -----------------------------------------------------
    parser_make_atom_poscars = subparsers.add_parser(
        name="make_atom_poscars",
        description="Tools for generating POSCAR files for atom calculations.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['map'])

    parser_make_atom_poscars.add_argument(
        "-d", "--dirname", type=Path, default=Path.cwd(),
        help="Directory name where atom calculation directories are created.")
    parser_make_atom_poscars.add_argument(
        "-e", "--elements", type=Element, default=None, nargs="+",
        help="Element names. When not set, all atom directories are created")

    parser_make_atom_poscars.set_defaults(func=make_atom_poscars)

    # -- make_spin_decomposed_volumetric_files ---------------------------------
    parser_spin_decomposed_volumetric_files = subparsers.add_parser(
        name="spin_decomposed_volumetric_files",
        description="Tools for making spin decomposed volumetric files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['sdvf'])

    parser_spin_decomposed_volumetric_files.add_argument(
        "-c", "--chgcar", type=str, required=True,
        help="CHGCAR-type file name with data for spin-down channel.")
    parser_spin_decomposed_volumetric_files.set_defaults(
        func=make_spin_decomposed_volumetric_files)

    # -- make_light_weight_vol_data --------------------------------------------
    parser_light_weight_vol_data = subparsers.add_parser(
        name="light_weight_vol_data",
        description="Tools for making light-weigh volumetric data file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['lwvd'])

    parser_light_weight_vol_data.add_argument(
        "-v", "--volumetric_file", type=str, required=True,
        help="CHGCAR-type volumetric file name.")
    parser_light_weight_vol_data.add_argument(
        "-o", "--output_lw_volumetric_filename", type=Path,
        help="Created file name.")
    parser_light_weight_vol_data.add_argument(
        "-b", "--border_fractions", type=float, nargs="*",
        default=default_border_fractions,
        help="Fractions that are used for depicting isosurfaces.")
    parser_light_weight_vol_data.add_argument(
        "-out_vesta", "--output_vesta_filename", type=Path,
        help="Output VESTA file name including volumetric file name and "
             "ISURFS.")
    parser_light_weight_vol_data.add_argument(
        "-orig_vesta", "--original_vesta_file", type=Path,
        help="Original VESTA file name to which volumetric file name and "
             "ISURFS are inserted.")

    parser_light_weight_vol_data.set_defaults(
        func=make_light_weight_vol_data)

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()




import os

from setuptools import setup, find_packages
from distutils.extension import Extension

from vise import __version__

module_dir = os.path.dirname(os.path.abspath(__file__))
reqs_raw = open(os.path.join(module_dir, "requirements.txt")).read()
reqs_list = [r.replace("==", "~=") for r in reqs_raw.split("\n")]
#reqs_list = [r for r in reqs_raw.split("\n")]

# try:
#     from Cython.Distutils import build_ext
# except ImportError:
#     use_cython = False
# else:
#     use_cython = True

cmdclass = {}
ext_modules = []

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='vise',
    version=__version__,
    author="Yu Kumagai",
    author_email="yukumagai@tohoku.ac.jp",
    url='https://github.com/kumagai-group/vise',
    packages=find_packages(),
    license='MIT license',
    description="Package for handling io of vasp package in kumagai group at "
                "IMR, Tohoku university",
    long_description=long_description,
    long_description_content_type="text/markdown",

    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License"
    ],
    python_requires='>=3.7',
    install_requires=reqs_list,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'vise = vise.cli.main:main',
            'vise_util = vise.cli.main_util:main',
        ]
    }
)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import tempfile

import pytest
from vise.input_set.prior_info import PriorInfo, prior_info_from_calc_dir
from vise.tests.helpers.assertion import assert_msonable


@pytest.fixture
def nonmagnetic_insulator():
    return PriorInfo(energy_per_atom=-0.5,
                     band_gap=1.2,
                     vbm_cbm=[1.0, 2.0],
                     total_magnetization=0.000001,
                     data_source="Materials Project",
                     is_cluster=False,
                     charge=2,
                     incar={"NUPDOWN": 2},
                     icsd_ids=[12, 13])


def test_round_trip_dict(nonmagnetic_insulator):
    assert_msonable(nonmagnetic_insulator)


def test_round_trip_yaml(nonmagnetic_insulator):
    tmp_file = tempfile.NamedTemporaryFile()
    nonmagnetic_insulator.dump_yaml(tmp_file.name)
    prior_info_from_json = PriorInfo.load_yaml(tmp_file.name)
    assert prior_info_from_json.as_dict() == nonmagnetic_insulator.as_dict()


def test_round_trip_json(nonmagnetic_insulator):
    tmp_file = tempfile.NamedTemporaryFile()
    nonmagnetic_insulator.dump_json(tmp_file.name)
    prior_info_from_json = PriorInfo.load_json(tmp_file.name)
    assert prior_info_from_json.as_dict() == nonmagnetic_insulator.as_dict()


def test_properties(nonmagnetic_insulator):
    assert nonmagnetic_insulator.is_magnetic is False
    assert nonmagnetic_insulator.has_band_gap is True
    assert nonmagnetic_insulator.is_metal is False


def test_input_options_kwargs(nonmagnetic_insulator):
    actual = nonmagnetic_insulator.input_options_kwargs
    expected = {"vbm_cbm": [1.0, 2.0],
                "is_magnetization": False,
                "band_gap": 1.2,
                "charge": 2}
    assert actual == expected


# def test_get_structure_from_prev_dir_actual_files(test_data_files):
#     prior_info = prior_info_from_calc_dir(prev_dir_path=test_data_files,
#                                           vasprun="MnO_uniform_vasprun.xml",
#                                           outcar="MnO_uniform_OUTCAR",
#                                           potcar="MnO_fake_POTCAR")
#     # Number of electrons in MnO_uniform_vasprun.xml is changed from 13 to 12
#     # artificially for this unittest.
#     assert prior_info.energy_per_atom == -8.024678125
#     assert pytest.approx(prior_info.band_gap) == 0.4702
#     assert prior_info.vbm_cbm == [4.6666, 5.1368]
#     assert prior_info.total_magnetization == 5.0000019
#     assert prior_info.charge == 1





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from typing import Optional, List

from monty.json import MSONable
from vise.input_set.kpoints_mode import KpointsMode
from vise.util.enum import ExtendedEnum


class Task(MSONable, ExtendedEnum):
    structure_opt = "structure_opt"
    structure_opt_rough = "structure_opt_rough"
    structure_opt_tight = "structure_opt_tight"
    cluster_opt = "cluster_opt"
    phonon_force = "phonon_force"
    defect = "defect"
    defect_2d = "defect_2d"
    band = "band"
    dos = "dos"
    dielectric_dfpt = "dielectric_dfpt"
    dielectric_finite_field = "dielectric_finite_field"
    dielectric_function = "dielectric_function"

    @property
    def is_lattice_relaxed(self) -> bool:
        return "structure_opt" in self.name

    @property
    def is_atom_relaxed_lattice_fixed(self) -> bool:
        return self in (self.cluster_opt, self.defect, self.defect_2d)

    @property
    def is_atom_relaxed(self) -> bool:
        return (self.is_lattice_relaxed
                or self.is_atom_relaxed_lattice_fixed)

    @property
    def is_dielectric(self) -> bool:
        return "dielectric" in self.name

    @property
    def is_tight_calc(self) -> bool:
        return self in (self.structure_opt_tight, self.phonon_force)

    @property
    def is_plot_task(self) -> bool:
        return self in (self.band, self.dos, self.dielectric_function)

    @property
    def is_spectrum_task(self) -> bool:
        return self in (self.dos, self.dielectric_function)

    @property
    def default_kpt_factor(self) -> Optional[int]:
        if self is self.dielectric_function:
            return 3
        elif self in (self.dos,
                      self.dielectric_dfpt,
                      self.dielectric_finite_field):
            return 2
        elif self.condition_kpt_factor_is_one:
            return 1
        else:
            raise NotImplementedError

    @property
    def condition_kpt_factor_is_one(self):
        if self in (self.band,
                    self.cluster_opt,
                    self.phonon_force,
                    self.defect,
                    self.defect_2d) or \
                self.is_lattice_relaxed:
            return True
        else:
            return False

    @property
    def default_kpt_mode(self) -> KpointsMode:
        if self == self.band:
            return KpointsMode.band
        elif self in (self.defect, self.defect_2d, self.cluster_opt,
                      self.phonon_force):
            return KpointsMode.uniform
        elif self.condition_kpoints_mode_is_primitive:
            return KpointsMode.primitive
        else:
            raise NotImplementedError

    @property
    def condition_kpoints_mode_is_primitive(self):
        if self is self.dos or \
                self.is_dielectric or \
                self.is_lattice_relaxed:
            return True
        else:
            return False

    @property
    def requisite_num_kpt_list(self) -> Optional[List[int]]:
        if self == self.cluster_opt:
            return [1, 1, 1]
        return

    @property
    def requisite_only_even_num_kpts(self) -> Optional[bool]:
        if self == self.cluster_opt:
            return False
        elif self == self.band:
            return True
        return

    # Gamma-center mesh is a must for GW calculations due to vasp
    # implementation and tetrahedron method, while is a strong recommend for
    # dos and dielectric function to sample the band edges.
    @property
    def requisite_gamma_centered(self) -> Optional[bool]:
        if self in (self.cluster_opt, self.phonon_force) \
                or self.is_spectrum_task:
            return True
        return

    @property
    def need_spin(self):
        return True if self in (self.defect, self.defect_2d) else False

    @property
    def fine_to_inherit_structure_from_prev(self):
        # For phonon_calculations, the POSCAR must no be inherited.
        return self is not self.phonon_force



# -*- coding: utf-8 -*-

import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, List

import yaml
from monty.json import MSONable
from monty.serialization import loadfn
from pymatgen.core import Structure
from pymatgen.io.vasp import Vasprun, Outcar, Potcar
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties
from vise.defaults import defaults


@dataclass
class PriorInfo(MSONable):
    """Used to control the input parameters"""
    structure: Structure = None
    energy_per_atom: float = None
    band_gap: float = None
    vbm_cbm: list = field(default_factory=list)
    total_magnetization: float = None
    data_source: str = None
    is_cluster: bool = None
    charge: int = None
    icsd_ids: List[int] = None
    incar: dict = field(default_factory=dict)

    def dump_yaml(self, filename: str = "prior_info.yaml") -> None:
        with open(filename, "w") as f:
            f.write(yaml.dump(self.as_dict()))

    @classmethod
    def load_yaml(cls, filename: str = "prior_info.yaml"):
        with open(filename, "r") as f:
            d = yaml.load(f, Loader=yaml.SafeLoader)

        return cls.from_dict(d)

    def dump_json(self, filename: str = "prior_info.json") -> None:
        with open(filename, "w") as fw:
            json.dump(self.as_dict(), fw, indent=2)

    @classmethod
    def load_json(cls, filename: str = "prior_info.json"):
        return loadfn(filename)

    @property
    def is_magnetic(self) -> Optional[bool]:
        try:
            return abs(self.total_magnetization) > defaults.integer_criterion
        except TypeError:
            return

    @property
    def has_band_gap(self) -> bool:
        return self.band_gap > defaults.band_gap_criterion

    @property
    def is_metal(self) -> bool:
        return not self.has_band_gap

    @property
    def input_options_kwargs(self):
        result = {}
        if self.vbm_cbm:
            result["vbm_cbm"] = self.vbm_cbm
        if isinstance(self.is_magnetic, bool):
            result["time_reversal"] = not self.is_magnetic
        if self.band_gap:
            result["band_gap"] = self.band_gap
        if self.charge:
            result["charge"] = self.charge
        return result


def prior_info_from_calc_dir(prev_dir_path: Path,
                             vasprun: str = "vasprun.xml",
                             outcar: str = "OUTCAR",
                             potcar: str = "POTCAR"):

    vasprun = Vasprun(str(prev_dir_path / vasprun))
    outcar = Outcar(str(prev_dir_path / outcar))
    potcar = Potcar.from_file(str(prev_dir_path / potcar))

    charge = get_net_charge_from_vasp(vasprun.final_structure,
                                      vasprun.parameters["NELECT"],
                                      potcar)
    structure = vasprun.final_structure.copy()
    energy_per_atom = outcar.final_energy / len(structure)
    band_edge_property = VaspBandEdgeProperties(vasprun, outcar)
    total_magnetization = outcar.total_mag

    return PriorInfo(structure=structure,
                     charge=charge,
                     energy_per_atom=energy_per_atom,
                     band_gap=band_edge_property.band_gap,
                     vbm_cbm=band_edge_property.vbm_cbm,
                     total_magnetization=total_magnetization)


def get_net_charge_from_vasp(structure: Structure,
                             nelect: int,
                             potcar: Potcar):
    """
    Returns the defect charge by comparing nion, number of electrons in POTCAR,
    and NELECT in INCAR.
    """
    nuclei_charge = 0

    for elem, potcar in zip(structure.composition, potcar):
        if potcar.element != str(elem):
            raise ValueError("The sequence of elements in POTCAR and Structure "
                             "is different.")
        nuclei_charge += potcar.nelectrons * structure.composition[elem]

    # charge is minus of difference of the electrons
    return int(nuclei_charge - nelect)



# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from phonopy.interface.calculator import read_crystal_structure
from phonopy.structure.atoms import PhonopyAtoms
from vise.util.phonopy.phonopy_input import structure_to_phonopy_atoms
import numpy as np


def assert_same_phonopy_atoms(actual: PhonopyAtoms,
                              expected: PhonopyAtoms):
    assert (actual.get_cell() == expected.get_cell()).all()
    assert (actual.get_scaled_positions()
            == expected.get_scaled_positions()).all()
    assert actual.symbols == expected.symbols


def test_phonopy_atoms_behavior(sc_structure, tmpdir):
    print(tmpdir)
    tmpdir.chdir()
#    actual = structure_to_phonopy_atoms(sc_structure)
    sc_structure.to(fmt="poscar", filename="POSCAR")
    a, _ = read_crystal_structure("POSCAR")
    b = PhonopyAtoms(atoms=a)
    print(type(a.get_cell()))
    print(a.get_atomic_numbers())
    assert_same_phonopy_atoms(a, b)


def test_structure_to_phonopy_atoms(sc_structure):
    actual = structure_to_phonopy_atoms(sc_structure)
    expected = PhonopyAtoms(symbols=["H"],
                            cell=np.array([[1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 0.0, 1.0]]),
                            scaled_positions=np.array([[0.0, 0.0, 0.0]]))
    assert_same_phonopy_atoms(actual, expected)

#
# def test_make_phonopy_input(mc_structure, mc_structure_conv):
#     actual = make_phonopy_input(unitcell=mc_structure,
#                                 supercell_matrix=np.eye(3).tolist(),
#                                 conventional_base=True)
#     supercell_matrix = [[ 1., 1., 0.],
#                         [-1., 1., 0.],
#                         [ 0., 0., 1.]]
#     supercell = mc_structure * supercell_matrix
#     expected = PhonopyInput(unitcell=mc_structure,
#                             supercell=supercell,
#                             supercell_matrix=supercell_matrix)
#     assert actual == expected
#
#
# def test_make_phonopy_input_default(mc_structure, mc_structure_conv):
#     actual = make_phonopy_input(unitcell=mc_structure)
#     supercell_matrix = [[ 2., 2., 0.],
#                         [-2., 2., 0.],
#                         [ 0., 0., 2.]]
#     supercell = mc_structure * supercell_matrix
#     expected = PhonopyInput(unitcell=mc_structure,
#                             supercell=supercell,
#                             supercell_matrix=supercell_matrix)
#     assert actual == expected
#
#
# def test_make_phonopy_input_default_hexa():
#     structure = Structure(Lattice.hexagonal(1.0, 2.0), species=["H"],
#                           coords=[[0.0]*3])
#     actual = make_phonopy_input(unitcell=structure)
#     supercell_matrix = [[2, -1, 0], [2, 1, 0], [0, 0, 2]]
#     supercell = structure * supercell_matrix
#     expected = PhonopyInput(unitcell=structure,
#                             supercell=supercell,
#                             supercell_matrix=supercell_matrix)
#     assert actual == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from copy import deepcopy, copy
from inspect import getfullargspec
from typing import Dict, Any, List

from pymatgen.core.structure import Structure

from vise import __version__
from vise.analyzer.band_edge_properties import is_band_gap
from vise.defaults import defaults
from vise.input_set.incar_settings_generator import IncarSettingsGenerator
from vise.input_set.potcar_generator import generate_potcar
from vise.input_set.structure_kpoints_generator import StructureKpointsGenerator
from vise.input_set.task import Task
from vise.input_set.vise_log import ViseLog
from vise.input_set.xc import Xc
from vise.util.logger import get_logger

logger = get_logger(__name__)


def function_args(func) -> List[str]:
    return getfullargspec(func).args


def constructor_args(target_class) -> List[str]:
    result = function_args(target_class.__init__)
    result.remove("self")
    return result


potcar_args = function_args(generate_potcar)
incar_settings_args = constructor_args(IncarSettingsGenerator)
structure_kpoints_args = constructor_args(StructureKpointsGenerator)

assignable_option_set = \
    set(structure_kpoints_args) | set(potcar_args) | set(incar_settings_args)


class CategorizedInputOptions:
    def __init__(self,
                 structure: Structure,
                 task: Task,
                 xc: Xc,
                 **input_options):

        self._input_options = deepcopy(input_options)
        self.initial_structure = structure.copy()
        self.task = task
        self.xc = xc
        self._set_kpt_density()
        self._raise_error_when_unknown_options_exist()

    def _raise_error_when_unknown_options_exist(self) -> None:
        unknown_args_set = self.input_option_set - assignable_option_set
        if unknown_args_set:
            raise ViseInputOptionsError(
                f"Options {unknown_args_set} are invalid")

    def _set_kpt_density(self):
        band_gap = self._input_options.get("band_gap", None)
        vbm_cbm = self._input_options.get("vbm_cbm", None)
        kpt_density = self._input_options.get("kpt_density", None)

        msg = ""
        if kpt_density is None:
            if self.task in (Task.defect, Task.defect_2d):
                kpt_density = defaults.defect_kpoint_density
                msg = " because task is defect"
            elif is_band_gap(band_gap, vbm_cbm):
                kpt_density = defaults.insulator_kpoint_density
                msg = " because there is a band gap"
            else:
                kpt_density = defaults.kpoint_density

            self._input_options["kpt_density"] = kpt_density

        logger.info(f"K-point density is set to {kpt_density}{msg}.")

    @property
    def all_options(self):
        result = copy(self._input_options)
        result.update({"task": self.task, "xc": self.xc,
                       "initial_structure": self.initial_structure})
        return result

    @property
    def input_option_set(self) -> set:
        return set(self._input_options.keys())

    def pick_target_options(self, target_args: List[str]) -> Dict[str, Any]:
        result = {}
        for target_arg in target_args:
            if target_arg in self.all_options:
                result[target_arg] = self.all_options[target_arg]
        return result

    @property
    def structure_kpoints_options(self) -> Dict[str, Any]:
        return self.pick_target_options(target_args=structure_kpoints_args)

    @property
    def potcar_options(self) -> Dict[str, Any]:
        return self.pick_target_options(target_args=potcar_args)

    @property
    def incar_settings_options(self) -> Dict[str, Any]:
        return self.pick_target_options(target_args=incar_settings_args)

    @property
    def vise_log(self):
        return ViseLog(version=__version__,
                       task=self.task, xc=self.xc,
                       input_options=self._input_options)


class ViseInputOptionsError(KeyError):
    pass




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from monty.json import MSONable
from monty.serialization import loadfn
from vise.util.mix_in import ToJsonFileMixIn, ToYamlFileMixIn, ToCsvFileMixIn


@dataclass
class TestTest(MSONable, ToJsonFileMixIn, ToCsvFileMixIn, ToYamlFileMixIn):
    a: str = "aaa"

    def as_dict(self):
        return {"a": self.a}

    # @classmethod
    # def from_yaml(cls, filename):
    #     d = loadfn(filename)
    #     return cls(**d)

    @property
    def to_dataframe(self):
        return pd.DataFrame({"a": [self.a]})

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame):
        d = df.to_dict()
        return cls.from_dict({"a": d["a"][0]})


def test_to_json_file_mix_in(tmpdir):
    actual = TestTest()
    tmpdir.chdir()
    actual.to_json_file()
    actual = loadfn("test_test.json")
    assert actual == TestTest()


def test_to_csv_file_mix_in(tmpdir):
    test = TestTest()
    tmpdir.chdir()
    test.to_csv_file()
    actual = Path("test_test.csv").read_text()
    expected = """a
aaa
"""
    assert actual == expected

    actual = TestTest.from_csv_file("test_test.csv")
    assert actual == test


def test_to_yaml_file_mix_in(tmpdir):
    actual = TestTest()
    tmpdir.chdir()
    print(tmpdir)
    actual.to_yaml_file()
    actual = TestTest.from_yaml("test_test.yaml")
    assert actual == TestTest()




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
# This file is originally developed by Naoki Tsunoda.

import itertools
from pathlib import Path
from typing import Union, Optional, Dict, Iterable, List, Tuple

import numpy as np

from vise.analyzer.vesta.element_colors import atom_color
from pymatgen.core import DummySpecies
from pymatgen.core.structure import Structure
from vise.util.logger import get_logger
from vise.util.unit_conversion import au_to_angstrom

logger = get_logger(__name__)


def replace_dummy_to_xx(target_str: str) -> str:
    return target_str.replace("X0+", "XX", -1)


def val_to_str_line(val_line: list or tuple) -> str:
    format_str = "{{:.{0}f}}".format(6)
    return " ".join([format_str.format(c) for c in val_line])


class VestaFile:
    def __init__(self,
                 structure: Structure,
                 title: str = None,
                 vectors: Dict[int, Iterable] = None,
                 vector_colors: List[Tuple[int, int, int]] = None,
                 bond_radius: float = 0.12,
                 boundary: Optional[Iterable] = None,
                 show_label: bool = True):
        """
        If name is set to the Site in structure, they are used for labels.

        Args:
            vectors (dict):
                e.g., {1: [1.0, 0.0, 0.0], ..}
            boundary (Iterable):
                a_min, a_max, b_min, b_max, c_min, c_max
                Center of origin is determined from the mean of boundary.
                e.g., "-0.5 0.5 -0.5 0.5 -0.5 0.5" means (0, 0, 0) as center.
        """
        self.blocks = [Title(structure, title),
                       Cellp(structure),
                       Struc(structure),
                       Bound(boundary),
                       SBond(structure, bond_radius=bond_radius),
                       SiteT(structure, show_label=show_label),
#                       DummyAtomt(structure),
                       Vect(vectors, size=0.2, colors=vector_colors),
                       Style(bond_radius=bond_radius)]

    def __repr__(self):
        outs = []
        for block in self.blocks:
            outs.append(repr(block))
        return "\n\n".join(outs)

    def write_file(self, filename: Union[Path, str]):
        filename = str(filename)
        filename = filename if ".vesta" in filename else f"{filename}.vesta"
        with open(filename, 'w') as poscar_vesta:
            poscar_vesta.write(repr(self))
        logger.info(f"{filename} is created.")


class Title:
    """ TITLE block in *.vesta files showing title in VESTA. """
    vesta_version = "#VESTA_FORMAT_VERSION 3.5.0"
    header = "TITLE"

    def __init__(self, structure: Structure, title: str = None):
        self.title = title or structure.formula

    def __repr__(self):
        return "\n".join([f'{self.vesta_version}',
                          f'{self.header}',
                          f'{self.title}'])


class Cellp:
    """
    CELLP block in *.vesta files, describing lattice constants.
    # e.g.) 3.992878   3.992878   3.992878  90.000000  90.000000  90.000000
    """
    header = "CELLP"

    def __init__(self, structure: Structure):
        self.cell_param = structure.lattice.lengths + structure.lattice.angles

    def __repr__(self):
        return "\n".join([self.header, val_to_str_line(self.cell_param)])


class Struc:
    """
    STRUC block in *.vesta files, showing atom sites.
    The index starts from 1
    e.g. "index" "element" "element+index" + "occupation" + "coords"
          1       Ba        Ba1               1.0000         0.500  0.500  0.500
          0 0 0 0 0
    """
    header = "STRUC"
    separator = " 0 0 0 0 0 "
    zero_coord = " 0.0 0.0 0.0 "

    def __init__(self, structure: Structure):
        coords = []
        for site_idx, site in enumerate(structure, 1):
            name = site.properties.get("name", None) \
                   or f'{site.species_string}{site_idx}'
            specifies = f'{site_idx} {site.species_string} {name} {1.0} '
            coord = val_to_str_line(site.frac_coords)
            coords.append(specifies + coord)
            coords.append(self.zero_coord)
        # replace "X0+" (pmg dummy species) -> "XX" (vesta dummy species)
        self.str_coords = replace_dummy_to_xx('\n'.join(coords))

    def __repr__(self):
        outs = [self.header, self.str_coords, self.separator]
        return '\n'.join(outs)


class ImportDensity:
    """
     IMPORT_DENSITY block in *.vesta files.
     """
    smooth_param = 0.1
    header = f"IMPORT_DENSITY {smooth_param}"
    prefix = "+1.000000"

    def __init__(self, volumetric_filename: str):
        """
        Args:
            volumetric_filename: e.g., 'CHGCAR'
        """
        self.string = f"{self.prefix} {volumetric_filename}"

    def __repr__(self):
        outs = [self.header, self.string]
        return "\n".join(outs)


class Bound:
    """control range of plot"""
    header = "BOUND"
    separator = " 0 0 0 0 0 "

    def __init__(self,
                 boundary: Optional[Union[tuple, list]] = (0, 1, 0, 1, 0, 1)):
        """
        Args:
            boundary (tuple or list): a_min, a_max, b_min, b_max, c_min, c_max
        """
        if boundary:
            if len(boundary) != 6:
                raise ValueError("length of boundary must be 6")
            self.boundary = val_to_str_line(boundary)
        else:
            self.boundary = None

    def __repr__(self):
        if self.boundary:
            outs = [self.header, self.boundary, self.separator]
            return "\n".join(outs)
        else:
            return ""


class SBond:
    """
    SBOND block in *.vesta files, control show/hide bonding between elements.

    boundary_mode = 0 : "Do not search the atoms beyond the boundary"
    """
    header = "SBOND"
    separator = " 0 0 0 0 "

    def __init__(self, structure: Structure, bond_factor=1.2,
                 boundary_mode=0, bond_radius: float = None):
        bond_list = []
        for i, (e1, e2) in enumerate(itertools.permutations(structure.types_of_species, 2), 1):
            try:
                bond = float(e1.average_ionic_radius + e2.average_ionic_radius) * bond_factor
            except AttributeError:
                continue
            x = f"{i} {e1.symbol} {e2.symbol} 0.0 {bond:5.4}  0  {boundary_mode}  1  0  1"
            if bond_radius:
                x += f" {bond_radius}   2.000 161  33 246"
            bond_list.append(x)
        self.bonds = "\n".join(bond_list)

    def __repr__(self):
        outs = [self.header, self.bonds, self.separator]
        return "\n".join(outs)


class SiteT:
    """
    SITET block in *.vesta files, control show/hide bonding between elements.

    boundary_mode = 0 : "Do not search the atoms beyond the boundary"
    """
    header = "SITET"
    separator = " 0 0 0 0 "

    def __init__(self, structure: Structure, show_label: bool = True):
        self.sites = []
        for i, site in enumerate(structure, 1):
            name = site.properties.get("name", None) or f'{site.species_string}{i}'
            if isinstance(site.specie, DummySpecies):
                rgb = "30 30 30"
            else:
                rgb = atom_color(site.species_string)
            radius = 0.3 if name == "center" else 0.5
            if show_label:
                label = 0 if name == "center" else 1
            else:
                label = 0
            self.sites.append(f" {i}  {name}  {radius} {rgb} {rgb} 204 {label}")

    def __repr__(self):
        outs = [self.header] + self.sites + [self.separator]
        return "\n".join(outs)


class Vect:
    """
     VECTR block in *.vesta files, showing directions of vectors
    e.g)
    1(vct_index)   -0.16519    0.00000    0.00000  <-- arrow end point
    1(atm_index)   0           0          0        <-- options

            penetration add_atomic_radius
    type 0:    no            no
    type 1:   yes            no
    type 2:    no           yes
    type 3:   yes           yes

    """
    header_1 = "VECTR"
    separator = " 0 0 0 0 0 "
    header_2 = "VECTT"
    type = 2

    def __init__(self,
                 vectors: Dict[int, Iterable],
                 size: float = 0.5,
                 colors: List[Tuple[int, int, int]] = None):
        """
        Args:
            vectors (dict): dict of site index and vector.
                            e.g., {1: [1.0, 0.0, 0.0], ..}

        RGB colors are in between 0 and 255
        (0, 0, 0) -> black
        """
        if vectors:
            vec_lines = []
            vector_types = []
            colors = colors or [(0, 0, 0)] * len(vectors)
            for i, (idx, vec) in enumerate(vectors.items(), 1):
                vec_lines.append(f'{i} {val_to_str_line(vec)}')
                vec_lines.append(f"{idx}  0 0 0 0")
                vec_lines.append(self.separator)
                c = f"{colors[i - 1][0]} {colors[i - 1][1]} {colors[i - 1][2]}"
                vector_option = f"{size} {c} {self.type}"
                vector_types.append(f'{i} {vector_option}')
            vec_lines.append(self.separator)
            vector_types.append(self.separator)
            self.vectors = '\n'.join(vec_lines)
            self.vector_types = '\n'.join(vector_types)
        else:
            self.vectors = None

    def __repr__(self):
        if self.vectors:
            return "\n".join([self.header_1, self.vectors, "",
                              self.header_2, self.vector_types])
        else:
            return ""


class DummyAtomt:
    """
    This is class object of ATOMT block in *.vesta files.
    Control style of atom object.
    (e.g., 1         Ba  0.8000  30 239  44  30 239  44 204)
    """
    header = "ATOMT"

    def __init__(self, structure: Structure):
        if "X" in structure.symbol_set:
            self.str = "  1         Xx  0.2000  76  76  76  76  76  76 504"
        else:
            self.str = ""

    def __repr__(self):
        return "\n".join([self.header, self.str]) if self.str else ""


class Style:
    """
    STYLE block in *.vesta files.
    """
    header = "STYLE"
    amp_prefix = "VECTS "
    atoms_prefix = "ATOMS "
    bondp_prefix = "BONDP"
    sects_prefix = "SECTS"
    sectp_prefix = "SECTP"
    ucolp_prefix = "UCOLP"

    all_bold_cell = "0  2  1.000   0   0   0"
    isurf = "1   1    12.0991 255 255   0 127 255"
    sect_param = 64
    sects = f" {sect_param}  1"
    sectp = "  1 0 0  0.00000E+00  0.00000E+00  -1.00000E+01  1.00000E+01"

    def __init__(self,
                 bond_radius: float,
                 is_ionic: bool = True):
        self.amplitude = 1.0
        self.atoms = "1  0  1" if is_ionic else "0  0  1"
        self.bond_radius = bond_radius

    def __repr__(self):
        vct = f'{self.amp_prefix} {self.amplitude}'
        sects = f'{self.sects_prefix} {self.sects}'
        sectp = f'{self.sectp_prefix} \n {self.sectp}'
        ucol = f'{self.ucolp_prefix} \n {self.all_bold_cell}'
        atoms = f'{self.atoms_prefix} {self.atoms}'
        bondp = f'{self.bondp_prefix} \n   1  16  {self.bond_radius}  1.000 127 127 127'
        return "\n".join([self.header, vct, sects, sectp, ucol, atoms, bondp])


def add_density(original_vesta_text: str,
                isurfs: List[float],
                volumetric_filename: str) -> str:
    """
    Note that when "CHG" is included in the volumetric file name, VESTA
    automatically divide the quantity by volume in cubic of atomic unit.
    """
    lines = original_vesta_text.split("\n")
    title_idx = lines.index("TITLE")
    # insertion timing must be here.
    import_density_line = ImportDensity(volumetric_filename).__repr__()
    lines.insert(title_idx + 3, import_density_line + "\n")

    lines.append("\nISURF")
    for isurf in isurfs:
        lines.append(f"  1   1  {isurf}  0  0  255  50  50")
    lines.append("  0   0   0   0""")
    return "\n".join(lines)


def calc_isurfs(border_fractions: List[float], is_chg: bool, volume: float
                ) -> List[float]:
    """Calc ISURFS values used in the VESTA files

    This is valid only for light-weighted volumetric data.
    """
    # Since max value is set to len(border_fracs), isurfs is multiplied by it.
    isurfs = np.array(border_fractions) * len(border_fractions)
    if is_chg:
        # VESTA uses atomic unit in length.
        isurfs /= (volume / (au_to_angstrom ** 3))
    return np.round(isurfs, 5).tolist()


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy
from dataclasses import dataclass
from typing import List, Dict, Optional, Tuple

import numpy as np
from monty.json import MSONable
from numpy import argmin
from pymatgen.electronic_structure.core import Spin

from vise.defaults import defaults
from vise.util.logger import get_logger

logger = get_logger(__name__)


@dataclass
class BandEdge(MSONable):
    energy: float
    spin: Spin
    band_index: int
    kpoint_coords: List[float]
    kpoint_index: int = None
    data_source: str = None
    symbol: str = None

    def is_direct(self, other: "BandEdge"):
        return (self.spin == other.spin and
                self.kpoint_coords == other.kpoint_coords)

    def __repr__(self):
        kpt_coords = f"{self.kpoint_coords[0]:5.3f} " \
                     f"{self.kpoint_coords[1]:5.3f} " \
                     f"{self.kpoint_coords[2]:5.3f}"
        return f"energy position: {self.energy}, spin: {self.spin.name:>4}, " \
               f"band index {self.band_index}, " \
               f"k-point index {self.kpoint_index}, k-point coords {kpt_coords}"

    def as_dict(self):
        return {"@module":       self.__class__.__module__,
                "@class":        self.__class__.__name__,
                "energy":        self.energy,
                "spin":          int(self.spin),
                "band_index":    self.band_index,
                "kpoint_index":  self.kpoint_index,
                "kpoint_coords": self.kpoint_coords,
                "data_source":   self.data_source,
                "symbol":        self.symbol}

    @classmethod
    def from_dict(cls, d):
        kwargs = deepcopy(d)
        kwargs["spin"] = Spin(d["spin"])
        kwargs.pop("@module", None)
        kwargs.pop("@class", None)
        return cls(**kwargs)


class BandEdgeProperties:
    def __init__(self,
                 eigenvalues: Dict[Spin, np.ndarray],
                 nelect: float,
                 magnetization: float,
                 kpoint_coords: List[List[float]],
                 integer_criterion: float = defaults.integer_criterion,
                 is_non_collinear: bool = False):

        assert 0 < integer_criterion < 0.5

        # [Spin][k-point_idx][band_idx] = eigenvalue
        self._eigenvalues = eigenvalues
        self._nelect = nelect
        #  In Bohr magneton.
        self._magnetization = magnetization
        # [[k1_x, k1_y, k1_z], [k2_x, k2_y, k2_z], ...]
        self._kpoint_coords = kpoint_coords
        self._integer_criterion = integer_criterion
        self._is_non_collinear = is_non_collinear

        self._calculate_vbm_cbm()

        if self._is_metal():
            self.vbm_info = None
            self.cbm_info = None

    def _calculate_vbm_cbm(self):
        vbm_energy = float("-inf")
        cbm_energy = float("inf")
        for spin, eigenvalues in self._eigenvalues.items():
            # ho = highest occupied
            ho_eigenvalue = np.amax(eigenvalues[:, self._ho_band_index(spin)])
            if ho_eigenvalue > vbm_energy:
                self.vbm_info = self.band_edge(
                    eigenvalues, self._ho_band_index(spin), ho_eigenvalue, spin)
                vbm_energy = ho_eigenvalue

            # lu = lowest unoccupied
            lu_band_index = self._ho_band_index(spin) + 1
            lu_eigenvalue = np.amin(eigenvalues[:, lu_band_index])
            if lu_eigenvalue < cbm_energy:
                self.cbm_info = self.band_edge(
                    eigenvalues, lu_band_index, lu_eigenvalue, spin)
                cbm_energy = lu_eigenvalue

    def band_edge(self, eigenvalues, band_index, eigenvalue, spin):
        k_index = int(np.where(eigenvalues[:, band_index] == eigenvalue)[0][0])
        return BandEdge(float(eigenvalue), spin, band_index,
                        self._kpoint_coords[k_index], k_index)

    def _ho_band_index(self, spin):
        if self._is_non_collinear:
            return round(self._nelect) - 1

        if spin == Spin.up:
            num_occupied_band = (self._nelect + self._magnetization) / 2
        else:
            num_occupied_band = (self._nelect - self._magnetization) / 2

        return round(num_occupied_band) - 1

    def _is_metal(self):
        nelect_frac = abs(round(self._nelect) - self._nelect)
        is_nelect_frac = nelect_frac > self._integer_criterion

        mag_frac = abs(round(self._magnetization) - self._magnetization)
        is_mag_frac = mag_frac > self._integer_criterion

        is_vbm_higher_than_cbm = self.vbm_info.energy > self.cbm_info.energy

        return is_nelect_frac or is_mag_frac or is_vbm_higher_than_cbm

    @property
    def is_metal(self):
        return self.vbm_info is None

    @property
    def is_direct(self):
        return self.vbm_info.is_direct(self.cbm_info) if self.vbm_info else None

    @property
    def band_gap(self):
        return self.cbm_info.energy - self.vbm_info.energy \
            if self.vbm_info else None

    def min_gap_w_coords(self, same_energy_threshold: float = 1e-5
                         ) -> Tuple[float, List[List[float]]]:
        """The direct-type minimum gap and its positions in the reciprocal space

        Note that the minimum positions can be multiple.

        Returns:
            min_gap (float): Minimum gap.
            kpoint_coords (List[List[float]]):
               positions in the reciprocal space in frac coords.
        """
        min_gap = float("inf")
        kpoint_indices = []
        for spin, eigenvalues in self._eigenvalues.items():
            lu_band_index = self._ho_band_index(spin) + 1

            ho_eigs = eigenvalues[:, self._ho_band_index(spin)]
            lu_eigs = eigenvalues[:, lu_band_index]
            diff = lu_eigs - ho_eigs
            _kpt = argmin(diff)
            if diff[_kpt] < min_gap + same_energy_threshold:
                min_gap = diff[_kpt]
                kpoint_indices.append(_kpt)
        kpoint_coords = [self._kpoint_coords[i] for i in kpoint_indices]
        return min_gap, kpoint_coords

    @property
    def vbm_cbm(self) -> List[float]:
        if self.vbm_info:
            return [self.vbm_info.energy, self.cbm_info.energy]

    def __repr__(self):
        if self.vbm_info:
            lines = [f"Band gap {self.band_gap:5.3f} eV",
                     f"VBM {self.vbm_info}",
                     f"CBM {self.cbm_info}"]
            return "\n".join(lines)
        return "Metal"


def is_band_gap(band_gap: Optional[float],
                vbm_cbm: Optional[List[float]],
                show_info: bool = False) -> bool:
    if not band_gap:
        band_gap = vbm_cbm[1] - vbm_cbm[0] if vbm_cbm else None
    if band_gap:
        band_gap_exist = band_gap > defaults.band_gap_criterion
        if show_info:
            x = ">" if band_gap_exist else "<"
            logger.info(
                f"Is there a band gap?: {band_gap_exist} " 
                f"({round(band_gap, 3)} {x} {defaults.band_gap_criterion} eV)")
        return band_gap_exist
    return False


def merge_band_edge(be_1: BandEdge, be_2: BandEdge, edge: str,
                    threshold=0.001) -> BandEdge:
    if edge not in ["vbm", "cbm"]:
        raise ValueError("edge needs to be vbm or cbm")

    if edge == "vbm":
        result = be_1 if be_1.energy > be_2.energy else be_2
    else:
        result = be_1 if be_1.energy < be_2.energy else be_2

    if abs(be_1.energy - be_2.energy) < threshold \
            and np.allclose(be_1.kpoint_coords, be_2.kpoint_coords) \
            and be_1.spin == be_2.spin:
        result.data_source = f"{be_1.data_source} {be_2.data_source}"

    return deepcopy(result)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from vise.util.bravais_lattice import BravaisLattice


@pytest.mark.parametrize("sg_num, bravais_lattice",
                         [(1, BravaisLattice.aP),
                          (3, BravaisLattice.mP),
                          (5, BravaisLattice.mC),
                          (16, BravaisLattice.oP),
                          (20, BravaisLattice.oC),
                          (22, BravaisLattice.oF),
                          (23, BravaisLattice.oI),
                          (38, BravaisLattice.oA),
                          (75, BravaisLattice.tP),
                          (79, BravaisLattice.tI),
                          (143, BravaisLattice.hP),
                          (146, BravaisLattice.hR),
                          (195, BravaisLattice.cP),
                          (196, BravaisLattice.cF),
                          (197, BravaisLattice.cI),
                          ])
def test_bravais_lattice_from_sg_num(sg_num, bravais_lattice):
    assert BravaisLattice.from_sg_num(sg_num) == bravais_lattice


def test_kpt_centering():
    assert BravaisLattice.oF.kpt_centering == [0.0, 0.0, 0.0]
    assert BravaisLattice.hP.kpt_centering == [0.0, 0.0, 0.5]
    assert BravaisLattice.cF.kpt_centering == [0.0, 0.0, 0.0]


def test_need_same_num_kpt():
    assert BravaisLattice.oI.need_same_num_kpt is True
    assert BravaisLattice.cP.need_same_num_kpt is False



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path
import numpy as np
from numpy.testing import assert_array_equal

import pytest

from pymatgen.electronic_structure.dos import CompleteDos, Dos
from pymatgen.core import Site
from pymatgen.electronic_structure.core import Spin, Orbital
from pymatgen.io.vasp import Vasprun

from vise.analyzer.vasp.dos_data import DosDataFromVasp
from vise.analyzer.plot_dos import DosPlotter


efermi = 0.5
energies = [-1, 0, 1]
tdos = [3, 4, 5]
pdos = [1, 2, 3]

total_dos = Dos(efermi=efermi, energies=energies,
                densities={Spin.up: tdos})
pdoses = {Site("H", [0, 0, 0]): {Orbital.s: {Spin.up: pdos},
                                 Orbital.px: {Spin.up: pdos},
                                 Orbital.py: {Spin.up: pdos},
                                 Orbital.pz: {Spin.up: pdos},
                                 Orbital.dxy: {Spin.up: pdos},
                                 Orbital.dyz: {Spin.up: pdos},
                                 Orbital.dxz: {Spin.up: pdos},
                                 Orbital.dz2: {Spin.up: pdos},
                                 Orbital.dx2: {Spin.up: pdos}}}


@pytest.fixture
def dos_data_from_vasp(mocker):
    stub_vasprun = mocker.Mock(spec=Vasprun)
    stub_vasprun.complete_dos = CompleteDos(None, total_dos, pdoses)
    return DosDataFromVasp(stub_vasprun).make_dos_data()


@pytest.fixture
def vasp_dos_data_crop_first_value(mocker):
    stub_vasprun = mocker.Mock(spec=Vasprun)
    stub_vasprun.complete_dos = CompleteDos(None, total_dos, pdoses)
    return DosDataFromVasp(stub_vasprun, crop_first_value=True).make_dos_data()


@pytest.fixture
def vasp_dos_data_energy_window(mocker):
    stub_vasprun = mocker.Mock(spec=Vasprun)
    stub_vasprun.complete_dos = CompleteDos(None, total_dos, pdoses)
    return DosDataFromVasp(stub_vasprun,
                           crop_first_value=True,
                           energy_window=[0.1, 1.1]).make_dos_data()


def test_energies(dos_data_from_vasp):
    assert dos_data_from_vasp.energies == energies


def test_total_dos(dos_data_from_vasp):
    assert_array_equal(dos_data_from_vasp.total, np.array([tdos]))


def test_pdos(dos_data_from_vasp):
    assert_array_equal(dos_data_from_vasp.pdos[0].s, np.array([pdos]))


def test_dos_for_crop_first_value(vasp_dos_data_crop_first_value):
    assert_array_equal(vasp_dos_data_crop_first_value.energies, energies[1:])
    assert_array_equal(vasp_dos_data_crop_first_value.total,
                       np.array([tdos[1:]]))
    assert_array_equal(vasp_dos_data_crop_first_value.pdos[0].s,
                       np.array([pdos[1:]]))


def test_energy_window(vasp_dos_data_energy_window):
    assert_array_equal(vasp_dos_data_energy_window.energies, energies[2])
    assert_array_equal(vasp_dos_data_energy_window.total,
                       np.array([tdos[2:]]))
    assert_array_equal(vasp_dos_data_energy_window.pdos[0].s,
                       np.array([pdos[2:]]))


def test_actual_vasp_files(test_data_files: Path):
    vasprun_file = str(test_data_files / "MgO_dos_vasprun.xml")
    vasprun = Vasprun(vasprun_file)
    dos_data = DosDataFromVasp(vasprun,
                               base_energy=1,
                               vertical_lines=[0.0]).make_dos_data()
    plot_data = dos_data.dos_plot_data({"Mg": [0], "O": [1]})
    plotter = DosPlotter(plot_data)
    plotter.construct_plot()
    plotter.plt.show()




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from copy import deepcopy
from dataclasses import dataclass
from itertools import cycle
from typing import List, Optional, Dict

import matplotlib.pyplot as plt
from monty.json import MSONable

from vise.error import ViseError
from vise.util.matplotlib import float_to_int_formatter
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.logger import get_logger

logger = get_logger(__name__)


@dataclass
class Irrep(MSONable):
    frac_coords: List[float]
    symbols: List[str]
    energies: List[float]
    degeneracies: List[int]

    @property
    def irrep_info_set(self):
        return zip(self.symbols, self.energies, self.degeneracies)


@dataclass
class Irreps(MSONable):
    sg_num: int
    # key is special point name. Gamma = GM
    irreps: Dict[str, Irrep]

    def __call__(self):
        return self.irreps

    def get_distances(self, x_ticks: "XTicks") -> List[List[float]]:
        result = []
        for special_point_symbol in self.irreps:
            dist_list = []
            for label, distance in zip(x_ticks.labels, x_ticks.distances):
                if special_point_symbol in label:
                    dist_list.append(distance)
            result.append(dist_list)
        return result


@dataclass(frozen=True)
class XTicks(MSONable):
    """Info used for X-axis ticks in the band structure plot.

    labels: Special point names, e.g., "Γ", "C$\mid$${\rm C}_2$",
    distances: Distances to the corresponding labels. So, the length must be the
    same as that of labels.
    """
    labels: List[str]
    distances: List[float]


@dataclass
class BandEdgeForPlot(MSONable):
    vbm: float
    cbm: float
    # The positions of the VBM and CBM denoted by total distances in the band
    # figure.
    vbm_distances: List[float]
    cbm_distances: List[float]


# backward compatibility
@dataclass
class BandEdge(BandEdgeForPlot):
    pass


@dataclass
class BandEnergyInfo(MSONable):
    # [branch][spin][band][k-point] A branch is an area in which
    # the k-points are continuous. Each branch is separated by a vertical bar.
    # We need to distinguish branch to draw continuous line in the area and
    # calculate the effective masses.
    band_energies: List[List[List[List[float]]]]
    band_edge: Optional[BandEdgeForPlot] = None
    fermi_level: Optional[float] = None
    irreps: Optional[Irreps] = None

    def __post_init__(self):
        if self.band_edge is None and self.fermi_level is None:
            raise ViseBandInfoError

    def slide_energies(self, base_energy):
        self._slide_band_energies(base_energy)
        self._slide_band_edge(base_energy)
        self._slide_fermi_level(base_energy)
        self._slide_irreps(base_energy)

    def _slide_band_energies(self, base_energy):
        self.band_energies = [[[[w - base_energy for w in x] for x in y]
                               for y in z] for z in self.band_energies]

    def _slide_band_edge(self, base_energy):
        if self.band_edge:
            self.band_edge.vbm -= base_energy
            self.band_edge.cbm -= base_energy

    def _slide_fermi_level(self, base_energy):
        if self.fermi_level:
            self.fermi_level -= base_energy

    def _slide_irreps(self, base_energy):
        if self.irreps:
            for i in self.irreps.irreps.values():
                i.energies = [e - base_energy for e in i.energies]

    @property
    def is_magnetic(self):
        return len(self.band_energies[0]) == 2

    def band_energy_region(self,
                           decision_width: float = 0.1,
                           bottom: float = None,
                           top: float = None,
                           offset: float = 0.0) -> List[List[float]]:
        """Estimate continuously filled band energy region.

        Args:
            decision_width: Max range over which energy is considered continuous
            bottom: minimum energy to be considered
            top: maximum energy to be considered
            offset: subtract offset value from all the band energy

        Returns:
            energy regions: [[1st min, 1st max], [2nd min, 2nd max], ...]
        """
        result = []

        def add_boundary(lower, upper):
            result.append([lower - offset, upper - offset])

        sorted_energies = sorted([energy
                                  for i in self.band_energies
                                  for j in i
                                  for k in j
                                  for energy in k])  # flatten nested list.
        if bottom is not None:
            sorted_energies = list(filter(lambda x: x >= bottom,
                                          sorted_energies))
        if top is not None:
            sorted_energies = list(filter(lambda x: x <= top,
                                          sorted_energies))

        prev_energy = sorted_energies.pop(0)
        lower_bound = prev_energy
        for energy in sorted_energies:
            if energy - prev_energy > decision_width:
                upper_bound = prev_energy
                add_boundary(lower_bound, upper_bound)
                lower_bound = energy  # update lower_bound
            prev_energy = energy
        else:
            upper_bound = energy  # last energy should be the upper bound.
            add_boundary(lower_bound, upper_bound)

        return result


@dataclass
class BandPlotInfo(MSONable, ToJsonFileMixIn):
    """Multiple BandInfo with the same k-point path are accepted.

    Ex: include both the PBE band and GW band
    """
    band_energy_infos: Dict[str, BandEnergyInfo]  # keys are subtitles.
    distances_by_branch: List[List[float]]
    x_ticks: XTicks
    title: str = None  # title of all plots

    def __post_init__(self):
        assert self.distances_by_branch[0][0] == self.x_ticks.distances[0]
        assert self.distances_by_branch[-1][-1] == self.x_ticks.distances[-1]

    def __add__(self, other: "BandPlotInfo"):
        assert self.distances_by_branch == other.distances_by_branch
        new_band_info_set = deepcopy(self.band_energy_infos)
        new_band_info_set.update(other.band_energy_infos)
        return BandPlotInfo(new_band_info_set, self.distances_by_branch,
                            self.x_ticks, self.title)


class ViseBandInfoError(ViseError):
    pass


class BandMplSettings:
    def __init__(self,
                 colors: Optional[List[str]] = None,
                 linewidth: List[float] = None,
                 circle_size: int = 70,
                 circle_colors: Optional[List[str]] = None,
                 band_edge_line_width: float = 0.75,
                 band_edge_line_color: str = "black",
                 band_edge_line_style: str = "-.",
                 title_font_size: int = 15,
                 label_font_size: int = 15,
                 show_legend: bool = True,
                 legend_location: str = "lower right"
                 ):
        self.colors = colors or ['#E15759', '#4E79A7', '#F28E2B', '#76B7B2']
        self.linewidth = cycle(linewidth) if linewidth else cycle([1.0])

        self.circle_size = circle_size
        self.hline = {"linewidth": band_edge_line_width,
                      "color": band_edge_line_color,
                      "linestyle": band_edge_line_style}

        self.title_font_size = title_font_size
        self.label_font_size = label_font_size

        self.circle_colors = circle_colors or ["pink", "green"]
        self.circle_size = circle_size

        self.show_legend = show_legend
        self.legend = {"loc": legend_location}

    def band_structure(self, index, label):
        result = {"color": self.colors[index],
                  "linewidth": next(self.linewidth),
                  "label": label}
        return result

    def circle(self, index):
        return {"color": self.colors[index],
                "marker": "o",
                "s": self.circle_size}


class BandMplPlotter:

    def __init__(self,
                 band_plot_info: BandPlotInfo,
                 energy_range: List[float],
                 base_energy: float = None,
                 base_energy_title: str = None,
                 mpl_defaults: Optional[BandMplSettings] = BandMplSettings()
                 ):
        # need deepcopy to avoid side effect caused by the energy shift
        band_plot_info = deepcopy(band_plot_info)
        self.band_energy_infos = band_plot_info.band_energy_infos
        self.distances_by_branch = band_plot_info.distances_by_branch
        self.x_ticks = band_plot_info.x_ticks

        self.energy_range = energy_range
        self.title = band_plot_info.title
        self.mpl_defaults = mpl_defaults
        self.plt = plt

        if base_energy is None:
            base_energy = get_base_energy(self.band_energy_infos, base_energy_title)
        slide_band_energies(self.band_energy_infos, base_energy)

    def construct_plot(self):
        self._add_band_set()
        self._set_figure_legend()
        self._set_x_range()
        self._set_y_range()
        self._set_labels()
        self._set_x_ticks()
        self._set_title()
        self._set_formatter()
        self.plt.tight_layout()

    def _add_band_set(self):
        for index, (band_name, band_info) in enumerate(self.band_energy_infos.items()):
            self._add_band_structures(band_info, band_name, index)

            if band_info.band_edge:
                self._add_band_edge(band_info.band_edge, index)
            else:
                if band_info.fermi_level:
                    self._add_fermi_level(band_info.fermi_level)

    def _add_band_structures(self, band_info, band_name, index):
        mpl_args = self.mpl_defaults.band_structure(index, band_name)
        for distances, energies_each_branch \
                in zip(self.distances_by_branch,  band_info.band_energies):
            for spin_index, energies_by_spin in enumerate(energies_each_branch):
                if spin_index == 0:
                    mpl_args.pop("linestyle", None)
                else:
                    mpl_args["linestyle"] = ":"

                for energies_of_a_band in energies_by_spin:
                    self.plt.plot(distances, energies_of_a_band, **mpl_args)
                    mpl_args.pop("label", None)

    def _add_band_edge(self, band_edge, index):
        self.plt.axhline(y=band_edge.vbm, **self.mpl_defaults.hline)
        self.plt.axhline(y=band_edge.cbm, **self.mpl_defaults.hline)

        for dist in band_edge.vbm_distances:
            self.plt.scatter(dist, band_edge.vbm,
                             **self.mpl_defaults.circle(index))
        for dist in band_edge.cbm_distances:
            self.plt.scatter(dist, band_edge.cbm,
                             **self.mpl_defaults.circle(index))

    def _add_fermi_level(self, fermi_level):
        self.plt.axhline(y=fermi_level, **self.mpl_defaults.hline)

    def _set_figure_legend(self):
        if self.mpl_defaults.show_legend and len(self.band_energy_infos) > 1:
            self.plt.legend(**self.mpl_defaults.legend)

    def _set_x_range(self):
        self.plt.xlim(self.x_ticks.distances[0], self.x_ticks.distances[-1])

    def _set_y_range(self):
        self.plt.ylim(self.energy_range[0], self.energy_range[1])

    def _set_labels(self):
        self.plt.xlabel("Wave vector", size=self.mpl_defaults.label_font_size)
        self.plt.ylabel("Energy (eV)", size=self.mpl_defaults.label_font_size)

    def _set_x_ticks(self):
        axis = self.plt.gca()
        axis.set_xticks(self.x_ticks.distances)
        axis.set_xticklabels(self.x_ticks.labels)
        for distance, label in zip(self.x_ticks.distances[1:-1],
                                   self.x_ticks.labels[1:-1]):
            linestyle = "-" if "\\mid" in label else "--"
            plt.axvline(x=distance, linestyle=linestyle)

    def _set_title(self):
        if self.title:
            self.plt.title(self.title, size=self.mpl_defaults.title_font_size)

    def _set_formatter(self):
        axis = self.plt.gca()
        axis.yaxis.set_major_formatter(float_to_int_formatter)


def get_base_energy(band_energy_infos, base_energy_title=None):
    if base_energy_title is None:
        base_energy_title = next(iter(band_energy_infos))
        if len(band_energy_infos) > 1:
            logger.warning(f"Base energy is set to {base_energy_title}.")

    base_band_info = band_energy_infos[base_energy_title]

    if base_band_info.band_edge:
        return base_band_info.band_edge.vbm

    return base_band_info.fermi_level


def slide_band_energies(band_energy_infos, base_energy):
    for band_info in band_energy_infos.values():
        band_info.slide_energies(base_energy)




# -*- coding: utf-8 -*-

from itertools import groupby
import operator
from typing import List

import numpy as np
from pymatgen.core import Structure

from vise.util.logger import get_logger


logger = get_logger(__name__)


def create_symbol_list(structure: Structure):
    species = [str(s) for s in structure.species]
    # unique_justseen https://docs.python.org/ja/3/library/itertools.html
    # ["H", "H", "O", "O", "H"] -> ['H', 'O', 'H']
    return list(map(next, map(operator.itemgetter(1), groupby(species, None))))


def sanitize_matrix(matrix: List[int]) -> List[List[int]]:

    if len(matrix) == 9:
        return [matrix[:3], matrix[3:6], matrix[6:]]
    elif len(matrix) == 3:
        result = np.eye(3, dtype=int)
        for i in range(3):
            result[i, i] = matrix[i]
        return result.tolist()
    elif len(matrix) == 1:
        result = np.eye(3, dtype=int)
        for i in range(3):
            result[i, i] = matrix[0]
        return result.tolist()
    else:
        raise ValueError(f"Matrix element length {len(matrix)} is improper.")


# -*- coding: utf-8 -*-
import json
from pathlib import Path

import pytest
from pymatgen.core import Lattice
from pymatgen.core.structure import Structure, IStructure
from vise.analyzer.plot_band import Irrep, Irreps
from vise.util.structure_symmetrizer import StructureSymmetrizer


@pytest.fixture(scope="session")
def test_data_files():
    return Path(__file__).parent / "test_data_files"


@pytest.fixture(scope="session")
def sc_structure():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0]]
    return Structure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def mc_structure():
    lattice =[[ 6.0, 3.0, 0.0],
              [ 6.0,-3.0, 0.0],
              [-4.0, 0.0, 7.0]]
    coords = [[0.0, 0.0, 0.0]]
    return Structure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def mc_structure_conv(mc_structure):
    symmetrizer = StructureSymmetrizer(mc_structure)
    return symmetrizer.conventional


@pytest.fixture(scope="session")
def complex_monoclinic_structure():
    coords = [[0.0, 0.0, 0.0],
              [0.1, 0.0, 0.0],
              [0.9, 0.0, 0.0],
              [0.2, 0.0, 0.0],
              [0.8, 0.0, 0.0],
              ]
    return Structure(lattice=Lattice.monoclinic(10, 20, 30, 75),
                     species=["H", "He", "He", "He", "He"],
                     coords=coords)


@pytest.fixture(scope="session")
def simple_cubic():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture
def irreps():
    irrep = Irrep(frac_coords=[0.0, 0.0, 0.0], symbols=["Γ1+"], energies=[0.1],
                  degeneracies=[1])
    return Irreps(sg_num=225, irreps={"Γ": irrep})

